"""Exec module for managing Rds Db Proxy Target Groups."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["soft_fail"]

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    resource_id: str,
    target_group_name="default",
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Args:
        resource_id(str): Name of the DB Proxy.

        target_group_name(str, Optional): Name of the DB Proxy Target Group.

        filters(List[dict[str, Any]], Optional): This parameter is not currently supported. Defaults to None.

            * Name (str): The name of the filter. Filter names are case-sensitive.

            * Values (List[str]): One or more filter values. Filter values are case-sensitive.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Resource State:

        .. code-block:: sls

            unmanaged_resource:
              exec.run:
                - path: aws.rds.db_proxy_target_group.get
                - kwargs:
                  resource_id: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_target_group.get resource_id=value
    """

    result = dict(comment=[], ret=None, result=True)

    # Read documentation: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html
    get_db_proxy_target_group = (
        await hub.exec.boto3.client.rds.describe_db_proxy_target_groups(
            ctx=ctx,
            **{
                "DBProxyName": resource_id,
                "TargetGroupName": target_group_name,
                "Filters": filters,
            },
        )
    )

    # Case: Error
    if not get_db_proxy_target_group["result"]:
        # Do not return success=false when it is not found.
        # Most of the resources would return "*NotFound*" type of exception when it is 404
        if "DBProxyTargetGroupNotFoundFault" in str(
            get_db_proxy_target_group["comment"]
        ) or "DBProxyNotFoundFault" in str(get_db_proxy_target_group["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_proxy_target_group", name=resource_id
                )
            )
            result["comment"].append(get_db_proxy_target_group["comment"])
            return result

        result["comment"].append(get_db_proxy_target_group["comment"])
        result["result"] = False
        return result

    # Case: Empty results
    if not get_db_proxy_target_group["ret"] or not get_db_proxy_target_group["ret"].get(
        "TargetGroups"
    ):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_proxy_target_group", name=resource_id
            )
        )
        return result

    raw_resource = get_db_proxy_target_group["ret"]["TargetGroups"][0]

    # Possible Resource attributes:
    result[
        "ret"
    ] = await hub.tool.aws.rds.db_proxy_target_group.convert_raw_resource_to_present_async(
        ctx=ctx,
        resource_id=resource_id,
        raw_resource=raw_resource,
        idem_resource_name=name,
    )

    return result


async def list_(
    hub,
    ctx,
    db_proxy_name: str,
    target_group_name="default",
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Args:
        db_proxy_name(str): Name of the DB Proxy.

        target_group_name(str, Optional): Name of the DB Proxy Target Group.

        filters(List[dict[str, Any]], Optional): This parameter is not currently supported. Defaults to None.

            * Name (str): The name of the filter. Filter names are case-sensitive.

            * Values (List[str]): One or more filter values. Filter values are case-sensitive.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        Resource State:

        .. code-block:: sls

            unmanaged_resources:
              exec.run:
                - path: aws.rds.db_proxy_target_group.list
                - kwargs:


        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_target_group.list

        Describe call from the CLI:

        .. code-block:: bash

            $ idem describe aws.rds.db_proxy_target_group

    """

    result = dict(comment=[], ret=[], result=True)

    # Read documentation: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html
    get_db_proxy_target_group = (
        await hub.exec.boto3.client.rds.describe_db_proxy_target_groups(
            ctx=ctx,
            **{
                "DBProxyName": db_proxy_name,
                "TargetGroupName": target_group_name,
                "Filters": filters,
            },
        )
    )

    if not get_db_proxy_target_group["result"]:
        result["comment"].append(get_db_proxy_target_group["comment"])
        result["result"] = False
        return result

    if not get_db_proxy_target_group["ret"].get("TargetGroups"):
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.rds.db_proxy_target_group", name=db_proxy_name
            )
        )
        return result

    # Possible Resource attributes:
    for resource in get_db_proxy_target_group["ret"]["TargetGroups"]:
        result["ret"].append(
            await hub.tool.aws.rds.db_proxy_target_group.convert_raw_resource_to_present_async(
                ctx=ctx,
                resource_id=db_proxy_name,
                raw_resource=resource,
                idem_resource_name=name if name else db_proxy_name,
            )
        )
    return result


async def update(
    hub,
    ctx,
    db_proxy_name: str,
    target_group_name: str = "default",
    connection_pool_config: make_dataclass(
        "ConnectionPoolConfiguration",
        [
            ("MaxConnectionsPercent", int, field(default=None)),
            ("MaxIdleConnectionsPercent", int, field(default=None)),
            ("ConnectionBorrowTimeout", int, field(default=None)),
            ("SessionPinningFilters", List[str], field(default=None)),
            ("InitQuery", str, field(default=None)),
        ],
    ) = None,
    resource_id: str = None,
    name: str = None,
) -> Dict[str, Any]:
    """
    Args:
        db_proxy_name(str): Name of the DB Proxy.

        target_group_name(str, Optional): Name of the DB Proxy Target Group

        connection_pool_config(Dict[str, Any], Optional): The settings that determine the size and behavior of the connection pool for the target group.
            * MaxConnectionsPercent (int) –
                The maximum size of the connection pool for each target in a target group. The value is expressed as a
                percentage of the max_connections setting for the RDS DB instance or Aurora DB cluster used by the target group.
                If you specify MaxIdleConnectionsPercent, then you must also include a value for this parameter.
                Default: 10 for RDS for Microsoft SQL Server, and 100 for all other engines
                Constraints: Must be between 1 and 100.

            * MaxIdleConnectionsPercent (int) –
                Controls how actively the proxy closes idle database connections in the connection pool.
                The value is expressed as a percentage of the max_connections setting for the RDS DB instance or
                Aurora DB cluster used by the target group. With a high value, the proxy leaves a high percentage of
                idle database connections open. A low value causes the proxy to close more idle connections and return
                them to the database.
                If you specify this parameter, then you must also include a value for MaxConnectionsPercent.
                Default: The default value is half of the value of MaxConnectionsPercent. For example,
                if MaxConnectionsPercent is 80, then the default value of MaxIdleConnectionsPercent is 40.
                If the value of MaxConnectionsPercent isn’t specified, then for SQL Server,
                MaxIdleConnectionsPercent is 5, and for all other engines, the default is 50.
                Constraints: Must be between 0 and the value of MaxConnectionsPercent.

            * ConnectionBorrowTimeout (integer):
                The number of seconds for a proxy to wait for a connection to become available in the connection pool.
                Only applies when the proxy has opened its maximum number of connections and all connections
                are busy with client sessions.
                Default: 120
                Constraints: between 1 and 3600, or 0 representing unlimited

            * SessionPinningFilters (List[str]):
                Each item in the list represents a class of SQL operations that normally cause all later statements in
                a session using a proxy to be pinned to the same underlying database connection. Including an item in
                the list exempts that class of SQL operations from the pinning behavior.
                Default: no session pinning filters

            * InitQuery (str):
                One or more SQL statements for the proxy to run when opening each new database connection.
                Typically used with SET statements to make sure that each connection has identical settings such as
                time zone and character set. For multiple statements, use semicolons as the separator.
                You can also include multiple variables in a single SET statement, such as SET x=1, y=2.
                Default: no initialization query

        resource_id(str, Optional): Name of the DB Proxy.

        name(str, Optional): Idem name of the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        Using in a state:

        .. code-block:: sls

            resource_is_present:
              aws.rds.db_proxy_target_group.present:
                - tags: value

        Exec call from the CLI:

        .. code-block:: bash

            idem exec aws.rds.db_proxy_target_group.create resource_id=value
    """

    result = dict(comment=[], ret={}, result=True)

    # Read documentation: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html
    modify_target_group = await hub.exec.boto3.client.rds.modify_db_proxy_target_group(
        ctx,
        **{
            "TargetGroupName": "default",
            "DBProxyName": db_proxy_name,
            "ConnectionPoolConfig": connection_pool_config,
        },
    )

    result["result"] = modify_target_group["result"]
    if not result["result"]:
        result["comment"].append(modify_target_group["comment"])
        return result

    result["comment"].append(
        f"Updated aws.rds.db_proxy_target_group '{name}'",
    )

    raw_resource = modify_target_group["ret"].get("DBProxyTargetGroup", {})
    result["ret"]["resource_id"] = raw_resource.get("DBProxyName")
    result["ret"]["name"] = name

    return result



"""Exec module for managing EC2 route table associations."""
from typing import Dict


async def get(
    hub,
    ctx,
    route_table_id: str,
    name: str = None,
    resource_id: str = None,
    gateway_id: str = None,
    subnet_id: str = None,
) -> Dict:
    """
    Get a route table association of a route table using Association ID, Gateway ID or Subnet ID.

    Args:
        route_table_id(str):
            ID of the AWS Route table.

        name(str, Optional):
            The name of the Idem state.

        resource_id(str, Optional):
            The route table association ID.

        gateway_id(str, Optional):
            The ID of the internet gateway or virtual private gateway.

        subnet_id(str, Optional):
            The ID of the subnet.


    Returns:
        Dict[str, Any]:
            Returns security group in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec aws.ec2.route_table_association.get name="my_resource" route_table_id=rtb-13ewdeae  resource_id=rtbassoc-0rera607rre12a2c5

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ec2.route_table_association.get
                - kwargs:
                    name: my_resource
                    route_table_id: rtb-13ewdeae
                    resource_id: rtbassoc-0rera607rre12a2c5
    """
    result = dict(comment=[], ret=None, result=True)

    route_table_ret = await hub.exec.boto3.client.ec2.describe_route_tables(
        ctx, RouteTableIds=[route_table_id]
    )

    if not route_table_ret["result"]:
        if "InvalidRouteTableID.NotFound" in str(route_table_ret.get("comment", "")):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.ec2.route_table_association", name=name
                )
            )
            result["result"] = False
            result["comment"] += list(route_table_ret["comment"])
            return result
        result["comment"] += list(route_table_ret.get("comment", ""))
        result["result"] = False
        return result

    if not route_table_ret.get("ret", {}).get("RouteTables"):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.route_table_association", name=name
            )
        )
        return result

    resource = route_table_ret["ret"].get("RouteTables")[0]
    association = hub.tool.aws.ec2.route_table.get_route_table_association_by_id(
        resource["Associations"],
        resource_id=resource_id,
        gateway_id=gateway_id,
        subnet_id=subnet_id,
    )

    if not association["result"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.ec2.route_table_association", name=name
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.aws.ec2.conversion_utils.convert_raw_route_table_association_to_present(
        resource=association["ret"],
        idem_resource_name=name,
    )
    return result



"""Exec module for managing Organization Configuration."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict


async def update(
    hub,
    ctx,
    resource_id: str,
    auto_enable: bool,
    data_sources: make_dataclass(
        """Describes which data sources will be enabled for the detector."""
        "DataSourceConfiguration",
        [
            (
                "S3Logs",
                make_dataclass(
                    """Describes whether S3 data event logs are enabled as a data source."""
                    "S3LogsConfiguration",
                    [("Enable", bool)],
                ),
                field(default=None),
            ),
            (
                "Kubernetes",
                make_dataclass(
                    """Describes whether any Kubernetes logs are enabled as data sources."""
                    "KubernetesConfiguration",
                    [
                        (
                            "AuditLogs",
                            make_dataclass(
                                """The status of Kubernetes audit logs as a data source."""
                                "KubernetesAuditLogsConfiguration",
                                [("Enable", bool)],
                            ),
                        )
                    ],
                ),
                field(default=None),
            ),
            (
                "MalwareProtection",
                make_dataclass(
                    """Describes whether Malware Protection is enabled as a data source."""
                    "MalwareProtectionConfiguration",
                    [
                        (
                            "ScanEc2InstanceWithFindings",
                            make_dataclass(
                                """Describes the configuration of Malware Protection for EC2 instances with findings."""
                                "ScanEc2InstanceWithFindingsConfiguration",
                                [("EbsVolumes", bool, field(default=None))],
                            ),
                            field(default=None),
                        )
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    org_conf=None,
) -> Dict[str, Any]:
    """Updates the delegated administrator account with the values provided.

    Args:
        resource_id(str):
            The ID of the detector to update the delegated administrator for.
        auto_enable(bool):
            Indicates whether to automatically enable member accounts in the organization.
        data_sources(dict, Optional):
            Describes which data sources will be updated.

            * S3Logs (*dict, Optional*):
                Describes whether S3 data event logs are enabled as a data source.

                * Enable (*bool*): The status of S3 data event logs as a data source.

            * Kubernetes (*dict, Optional*):
                Describes whether any Kubernetes logs are enabled as data sources.

                * AuditLogs (*dict*):
                    The status of Kubernetes audit logs as a data source.

                    * Enable (*bool*):
                        The status of Kubernetes audit logs as a data source.

            * MalwareProtection (*dict, Optional*):
                Describes whether Malware Protection is enabled as a data source.

                * ScanEc2InstanceWithFindings (*dict, Optional*):
                    Describes the configuration of Malware Protection for EC2 instances with findings.

                    EbsVolumes (*bool, Optional*):
                        Describes the configuration for scanning EBS volumes as data source.
        org_conf(Optional):
            Describes current state of Organization Configuration.


    Returns:
        Dict[str, Any]:
            Returns organization Configuration in updated format.

    Examples:
        Calling this exec module function from the cli.

        .. code-block:: yaml

            my_unmanaged_resources:
              exec.run:
                - path: aws.guardduty.organization_configuration.update
                - kwargs:
                    - resource_id: 'string'
                    - auto_enable: True|False
                    - data_sources:
                        S3Logs:
                            Enable: True|False
                        Kubernetes:
                            AuditLogs:
                                Enable: True|False
                        MalwareProtection:
                            ScanEc2InstanceWithFindings:
                                EbsVolumes: True|False
    """
    result = dict(comment=[], ret=None, result=True)

    if not org_conf:
        response = await get(hub, ctx, resource_id=resource_id)
    else:
        response = org_conf
    if response["ret"]:
        resource_updated = await hub.tool.aws.guardduty.config_utils.is_organization_configuration_updated(
            before=response["ret"],
            auto_enable=auto_enable,
            data_sources=data_sources,
        )
        if resource_updated:
            ret = (
                await hub.exec.boto3.client.guardduty.update_organization_configuration(
                    ctx,
                    DetectorId=resource_id,
                    AutoEnable=auto_enable,
                    DataSources=data_sources,
                )
            )
            if ret["result"]:
                result["comment"].append(
                    hub.tool.aws.comment_utils.update_comment(
                        resource_type="aws.guardduty.organization_configuration",
                        name=resource_id,
                    )
                )
                result["ret"] = ret["ret"]
            else:
                result["result"] = ret["result"]
                result["comment"].append(ret["comment"])
            return result
        else:
            result["comment"].append(
                hub.tool.aws.comment_utils.already_exists_comment(
                    resource_type="aws.guardduty.organization_configuration",
                    name=resource_id,
                )
            )
            return result
    else:
        result["result"] = False
        result["comment"] = response["comment"]
        return result


async def get(hub, ctx, resource_id: str) -> Dict:
    """Returns information about the account selected as the delegated administrator for GuardDuty.

    Args:
        resource_id(str):
            AWS Detector ID to identify the resource.

    Returns:
        Dict[str, Any]:
            Returns organization Configuration in updated format

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.guardduty.organization_configuration.get resource_id="detector_id"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.guardduty.organization_configuration.get
                - kwargs:
                    resource_id: "detector_id"
    """
    result = dict(comment=[], ret=None, result=True)
    organization_configuration = (
        await hub.exec.boto3.client.guardduty.describe_organization_configuration(
            ctx, DetectorId=resource_id
        )
    )
    if not organization_configuration["result"]:
        if "NoSuchEntity" in str(organization_configuration["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.guardduty.organization_configuration",
                    name=resource_id,
                )
            )
            result["comment"] += list(organization_configuration["comment"])
            return result
        result["result"] = False
        result["comment"] += list(organization_configuration["comment"])
        return result

    if not (organization_configuration["ret"]):
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.guardduty.organization_configuration",
                organization_configuration=organization_configuration,
                name=resource_id,
            )
        )
        return result

    result[
        "ret"
    ] = hub.tool.aws.guardduty.conversion_utils.convert_raw_organization_configuration_to_present(
        organization_configuration=organization_configuration,
        idem_resource_name=resource_id,
    )
    return result



"""
Functions to help with state arguments via pre contracts
"""
from typing import Any
from typing import Dict


def sync_sls_name_and_tag(hub, ref: str, kwargs: Dict[str, Any]):
    """
    Modify the kwargs passed to a state to have the same value for "Name" tag and sls name
    """
    # If the given state doesn't have "tags" as a parameter then quit early
    try:
        if "tags" not in hub[ref].present.signature.parameters:
            return
    except KeyError:
        return

    try:
        sync_sls_name_and_tag = hub.OPT.idem.sync_sls_name_and_name_tag
    except KeyError:
        sync_sls_name_and_tag = False

    if not sync_sls_name_and_tag:
        return

    # Make the "name" attribute and the "Name" tag match each other

    name = kwargs["name"]

    # Transform tag_specification to tags
    hub.tool.aws.parameters.tag_specs_to_dict(kwargs)
    tags = kwargs["tags"]

    # Patch the "name" attribute to match the "Name" tag
    kwargs["name"] = tags.get("Name", name)

    # Patch the "Name" tag to match the "name" attribute
    if "Name" not in tags:
        tags["Name"] = name

    kwargs["tags"] = tags


def tag_specs_to_dict(hub, kwargs: Dict[str, Any]):
    """
    Turn tag specificiations to a dict
    """
    tags = kwargs.get("tags") or {}

    # Transform tag_specification to tags
    if isinstance(tags, list):
        tags = {t["Key"]: t["Value"] for t in tags}

    # Override the "tags" in kwargs with the new value
    kwargs["tags"] = tags



"""
Exec module for rds.db_instance
"""
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__func_alias__ = {"list_": "list"}


async def get(hub, ctx, resource_id: str, name: str = None) -> Dict[str, Any]:
    """
    Returns information about provisioned RDS instances. This operation can also return information for Amazon Neptune
    DB instances and Amazon DocumentDB instances.

        Args:
            resource_id (str):
                An identifier of the resource in the provider. Defaults to None.

            name(str, Optional): Idem name of the resource. Defaults to None.

        Examples:

            .. code-block:: bash

                idem exec aws_auto.rds.db_instance.get resource_id=0123456789abcdef

            .. code-block:: python

                async def my_func(hub, ctx):
                    ret = await hub.exec.aws_auto.rds.db_instance.get(ctx, resource_id="01235789abcdef")

            .. code-block:: yaml

                aws_auto_rds_db_instance_get_resource:
                  exec.run:
                    - path: aws_auto.rds.db_instance.get
                    - resource_id: 0123456789abcdef

    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    ret = await hub.exec.boto3.client.rds.describe_db_instances(
        ctx=ctx,
        DBInstanceIdentifier=resource_id,
    )
    if not ret["result"]:
        if "DBInstanceNotFoundFault" in str(ret.get("comment", "")):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.rds.db_instance",
                    name=name if name else resource_id,
                )
            )
        else:
            ret["result"] = False
        result["comment"] += list(ret["comment"])
        return result
    if not ret["ret"]["DBInstances"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.rds.db_instance", name=name if name else resource_id
            )
        )
        return result

    resource = ret["ret"]["DBInstances"][0]
    if len(ret["ret"]["DBInstances"]) > 1:
        result["comment"].append(
            f"More than one aws.rds.db_instance resource was found. Use resource {resource.get('DBInstanceIdentifier')}"
        )
    tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
        ctx, ResourceName=resource["DBInstanceArn"]
    )
    if not tags["result"]:
        result["result"] = False
        result["comment"] = tags["comment"]
        return result
    result[
        "ret"
    ] = hub.tool.aws.rds.conversion_utils.convert_raw_db_instance_to_present(
        raw_resource=resource,
        raw_resource_tags=tags,
    )
    return result


async def list_(
    hub,
    ctx,
    filters: List[
        make_dataclass("Filter", [("Name", str), ("Values", List[str])])
    ] = None,
) -> Dict[str, Any]:
    """
    Returns information about provisioned RDS instances. This operation can also return information for Amazon
    Neptune DB instances and Amazon DocumentDB instances.

        Args:
            filters (list[Dict[str, Any]], Optional):
                A filter that specifies one or more DB instances to describe.

                Supported filters:
                * db-cluster-id (str, Optional):
                    Accepts DB cluster identifiers and DB cluster Amazon Resource Names (ARNs). The results list only
                    includes information about the DB instances associated with the DB clusters identified by these
                    ARNs.

                * db-instance-id (str, Optional):
                    Accepts DB instance identifiers and DB instance Amazon Resource Names (ARNs). The results list only
                    includes information about the DB instances identified by these ARNs.

                * dbi-resource-id (str, Optional):
                    Accepts DB instance resource identifiers. The results list will only include information about the
                    DB instances identified by these DB instance resource identifiers.

                * domain (str, Optional):
                    Accepts Active Directory directory IDs. The results list only includes information about the DB
                    instances associated with these domains.

                * engine (str, Optional):
                    Accepts engine names. The results list only includes information about the DB instances for these
                    engines. Defaults to None.

                * Name (str):
                    The name of the filter. Filter names are case-sensitive.

                * Values (list[str]):
                    One or more filter values. Filter values are case-sensitive.

        Examples:

            .. code-block:: bash

                idem exec aws_auto.rds.db_instance.list

            .. code-block:: python

                async def my_func(hub, ctx):
                    ret = await hub.exec.aws_auto.rds.db_instance.list(ctx)

            .. code-block:: yaml

                aws_auto_rds_db_instance_list_resource:
                  exec.run:
                    - path: aws_auto.rds.db_instance.list

    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }
    ret = await hub.exec.boto3.client.rds.describe_db_instances(
        ctx=ctx,
        filters=filters,
    )

    result["ret"] = []
    for resource in ret["ret"]["DBInstances"]:
        tags = await hub.exec.boto3.client.rds.list_tags_for_resource(
            ctx,
            ResourceName=resource["DBInstanceArn"],
        )
        if not tags["result"]:
            result["result"] = False
            result["comment"] = tags["comment"]
            return result
        result["ret"].append(
            hub.tool.aws.rds.conversion_utils.convert_raw_db_instance_to_present(
                raw_resource=resource,
                raw_resource_tags=tags,
            )
        )
    return result



"""Exec module for managing SES Domain Identity."""

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, resource_id: str):
    """Get SES domain identity from AWS account.

    Get a single domain identity from AWS.

    Args:
        resource_id (str):
            The domain identity.
    Returns:
        Dict[str, Any]:
            Returns domain identity in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.ses.domain_identity.get resource_id="example.com"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ses.domain_identity.get
                - kwargs:
                    resource_id:"example.com"
    """

    result = dict(comment=[], ret=None, result=True)

    ret = await hub.exec.aws.ses.domain_identity.list(ctx, [resource_id])

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    if len(ret["ret"]) != 0:
        result["ret"] = ret["ret"][0]

    return result


async def list_(hub, ctx, identities: list = None):
    """List SES domain identities from AWS account.

    Fetch a list of domain identities from AWS. The function returns empty list when no resource is found.

    Args:
        identities (str, Optional):
            The domain identities.
            If no identity is specified, it will return all the identities present in AWS account.

    Returns:
        Dict[str, Any]:
            Return domain identities in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.ses.domain_identity.list identity=["example.com"]

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.ses.domain_identity.list
                - kwargs:
                    identity:"example.com"
    """
    result = dict(comment=[], ret=[], result=True)

    if not identities:
        ret_list = await hub.exec.boto3.client.ses.list_identities(ctx)

        if not ret_list["result"]:
            result["comment"] += list(ret_list["comment"])
            result["result"] = False
            hub.log.debug(f"Could not list identities {ret_list['comment']}")
            return result

        if not ret_list.get("ret"):
            result["comment"] += ("No domain identities present",)
            return result
        identities = ret_list["ret"].get("Identities")

    ret = await hub.exec.boto3.client.ses.get_identity_verification_attributes(
        ctx, Identities=identities
    )

    if not ret["result"]:
        result["comment"] += list(ret["comment"])
        result["result"] = False
        return result

    identities = ret["ret"]["VerificationAttributes"]

    for identity in identities:
        result["ret"].append(
            hub.tool.aws.ses.conversion_utils.convert_raw_domain_identity_to_present(
                ctx,
                raw_resource=identity,
            )
        )

    return result



"""
Manage Ec2 Network Interfaces
"""
from typing import Any
from typing import Dict
from typing import List

from dict_tools.typing import Computed

__contracts__ = ["resource", "allow_sync_sls_name_and_name_tag"]  # , "soft_fail"]


TREQ = {
    "present": {
        "require": [
            "aws.ec2.subnet.present",
        ],
    },
}


async def present(
    hub,
    ctx,
    *,
    name: str,
    subnet_id: str,
    resource_id: str = None,
    client_token: str = None,
    description: str = None,
    groups: List[str] = None,
    interface_type: str = None,
    default_ipv6_address: str = None,
    primary_ip_address: str = None,
    private_ip_addresses: List[str] = None,
    ipv4_address_count: int = None,
    ipv4_prefix_count: int = None,
    ipv4_prefixes: List[str] = None,
    ipv6_addresses: List[str] = None,
    ipv6_address_count: int = None,
    ipv6_prefix_count: int = None,
    ipv6_prefixes: List[str] = None,
    tags: Dict[str, str] = None,
    # Parameters cannot be used on creation
    # Computed Parameters (cannot be modified) # TODO some of these CAN be modified
    availability_zone: Computed[str] = None,
    mac_address: Computed[str] = None,
    outpost_arn: Computed[str] = None,
    owner_id: Computed[str] = None,
    private_dns_name: Computed[str] = None,
    requester_id: Computed[str] = None,
    requester_managed: Computed[str] = None,
    source_dest_check: Computed[str] = None,
    status: Computed[str] = None,
    vpc_id: Computed[str] = None,
    deny_all_igw_traffic: Computed[bool] = None,
    ipv6_native: Computed[bool] = None,
    # Association
    allocation_id: Computed[str] = None,
    ip_owner_id: Computed[str] = None,
    public_dns_name: Computed[str] = None,
    public_ip: Computed[str] = None,
    customer_owned_ip: Computed[str] = None,
    carrier_ip: Computed[str] = None,
    # Attachment
    attachment_id: Computed[str] = None,
    delete_on_termination: Computed[bool] = None,
    device_index: Computed[int] = None,
    network_card_index: Computed[int] = None,
    ena_srd_enabled: Computed[bool] = None,
    ena_srd_udp_enabled: Computed[bool] = None,
    **kwargs,
) -> Dict[str, Any]:
    """
    Creates a network interface in the specified subnet. The number of IP addresses you can assign to a network
    interface varies by instance type. For more information, see IP Addresses Per ENI Per Instance Type in the
    Amazon Virtual Private Cloud User Guide. For more information about network interfaces, see Elastic network
    interfaces in the Amazon Elastic Compute Cloud User Guide.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str):
            An identifier of the resource in the provider. Defaults to None.

        client_token(str, Optional):
            The idempotency token for the network interface.

        description(str, Optional):
            A description for the network interface. Defaults to None.

        groups(List[str], Optional): The IDs of one or more security groups. Defaults to None.

        ipv6_addresses(List[str]], Optional):
            The IPv6 addresses from the IPv6 CIDR block range of your subnet.  Defaults to None.

        default_ipv6_address(str, Optional):
            The default IPv6 address.

        primary_ip_address(str, Optional):
            The primary private IPv4 address of the network interface. If you don't specify an IPv4 address,
            Amazon EC2 selects one for you from the subnet's IPv4 CIDR range. Defaults to None.

        private_ip_addresses(List[str]], Optional):
            The private IPv4 addresses. If you specify the private IPv4 addresses, you should also specify the
            primary private IPv4 address using the primary_ip_address property. If you only need a primary private
            IPv4 address, you can specify it using the primary_ip_address property and not use this property.

        interface_type(str, Optional):
            The type of network interface. The default is interface. The only supported values are efa and
            trunk. Defaults to None.

        subnet_id(str):
            The ID of the subnet to associate with the network interface.

        tags(Dict[str, str], Optional):
            The tags to apply to the resource. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            default_vpc:
              exec.run:
                - path: aws.ec2.vpc.get
                - kwargs:
                    filters:
                      - name: is-default
                        values:
                          - "true"

            default_subnet:
              exec.run:
                - path: aws.ec2.subnet.get
                - kwargs:
                    filters:
                      - name: vpc-id
                        values:
                          - ${exec:default_vpc:resource_id}

            my_network_interface:
              aws.ec2.network_interface.present:
                - subnet_id: ${exec:default_subnet:resource_id}
    """
    # Convert a list of dictionaries to a plain dictionary before calculating desired_state
    if isinstance(tags, list):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    # Get all the parameters passed to this function as a single dictionary
    desired_state = {
        k: v for k, v in locals().items() if k not in ("hub", "ctx", "kwargs")
    }

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    current_state = None
    if resource_id:
        get = await hub.exec.aws.ec2.network_interface.get(
            ctx, name=name, resource_id=resource_id
        )

        if get.result is True:
            if get.ret is None:
                result["comment"] += [
                    f"Could not find Network Interface for '{name}' with existing id '{resource_id}'"
                ]
                result["result"] = False
                return result
            else:
                result["comment"] += [f"Network Interface '{name}' already exists"]
                current_state = result["old_state"] = get.ret
    elif client_token:
        get = await hub.exec.aws.ec2.instance.get(
            ctx, name=name, filters=[{"Name": "client-token", "Values": [client_token]}]
        )

        if get.result is True:
            if get.ret is None:
                result["comment"] += [
                    f"Could not find Network Interface for '{name}' with existing idempotence token '{client_token}'"
                ]
            else:
                current_state = result["old_state"] = get.ret
                resource_id = current_state.resource_id
                result["comment"] += [f"Network Interface '{name}' already exists"]

    # No resource_id from cache or parameters, we need to create the resource
    if not resource_id:
        if ctx.test:
            result["new_state"] = desired_state
            result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.ec2.network_interface", name=name
            )
            return result
        else:
            create_ret = await hub.exec.aws.ec2.network_interface.create(
                ctx, **desired_state
            )
            # If we fail to create the network interface, just return the proper error message
            if not create_ret["result"]:
                result["result"] = False
                result["comment"] += [create_ret.comment]
                return result

            result["comment"] += hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.ec2.network_interface", name=name
            )
            result["result"] &= create_ret.result
            result["new_state"] = current_state = create_ret.ret
            # This makes sure the created Network Interface is saved to esm regardless if the subsequent update call fails or not.
            result["force_save"] = True

    if not current_state:
        result["comment"] += [f"Unable to get the current_state: '{name}'"]
        result["result"] = False
        return result

    # Modify all the attributes that need modification
    if ctx.test:
        result["new_state"] = desired_state
    else:
        ctx.old_state = current_state
        await hub.exec.aws.ec2.network_interface.update(ctx, **desired_state)
        get = await hub.exec.aws.ec2.network_interface.get(
            ctx, name=name, resource_id=current_state["resource_id"]
        )
        if get.result:
            result["new_state"] = get.ret

    return result


async def absent(
    hub, ctx, *, name: str, resource_id: str = None, client_token: str = None, **kwargs
) -> Dict[str, Any]:
    r"""
    Deletes the specified network interface. You must detach the network interface before you can delete it.

    Args:
        name(str):
        An Idem name of the resource.

        resource_id(str, Optional):
            The ID of the network interface.

        client_token(str, Optional):
            The idempotency token for the network interface.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.ec2.network_interface.absent:
                - name: value
                - resource_id: value
    """
    result = dict(
        comment=[], old_state=ctx.old_state, new_state=None, name=name, result=True
    )

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if not resource_id and client_token:
        get = await hub.exec.aws.ec2.instance.get(
            ctx, name=name, filters=[{"Name": "client-token", "Values": [client_token]}]
        )
        resource_id = get.ret.resource_id

    # If there still is no resource_id, the instance is gone
    if not resource_id:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.network_interface", name=name
        )
        return result

    if ctx.test:
        result["comment"] += hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.ec2.network_interface", name=name
        )
        return result

    ret = await hub.exec.aws.ec2.network_interface.delete(ctx, resource_id=resource_id)
    result["comment"].extend(ret.comment)

    if not ret.result and any("does not exist" in c for c in ret.comment):
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.ec2.network_interface", name=name
        )
        # The command failed because the instance no longer exists
        return result

    elif not ret.result:
        # The failure is some other reason we need to examine
        result["result"] &= ret.result
        result["comment"] += [ret.comment]
    else:
        result["comment"] += hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.ec2.network_interface", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Describes one or more of your network interfaces.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws_auto.ec2.network_interface
    """
    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_network_interfaces(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe network_interface {ret['comment']}")
        return result

    network_interfaces = hub.tool.aws.ec2.network_interface.convert_to_present(
        ret["ret"]
    )

    for resource_id, resource_present_state in network_interfaces.items():
        result[resource_id] = {
            "aws.ec2.network_interface.present": [
                {k: v} for k, v in resource_present_state.items() if v is not None
            ]
        }

    return result



"""State module for managing AWS Api Gateway Domain Name."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    certificate_name: str = None,
    certificate_arn: str = None,
    regional_certificate_name: str = None,
    regional_certificate_arn: str = None,
    resource_id: str = None,
    endpoint_configuration: make_dataclass(
        "endpointConfigurationInput",
        [
            ("types", List[str], field(default=None)),
            ("vpcEndpointIds", List[str], field(default=None)),
        ],
    ) = None,
    mutual_tls_authentication: make_dataclass(
        "MutualTlsAuthenticationInput",
        [
            ("TruststoreUri", str, field(default=None)),
            ("TruststoreVersion", str, field(default=None)),
        ],
    ) = None,
    security_policy: str = None,
    tags: Dict[str, str] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=30)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates or updates the apigateway domain name resource.

    Args:
        name(str):
            The name of the DomainName resource
        certificate_name(str, Optional):
            The user-friendly name of the certificate that will be used
            by edge-optimized endpoint for this domain name.
        certificate_arn(str, Optional):
            The reference to an AWS-managed certificate that will be used by
            edge-optimized endpoint for this domain name. AWS Certificate Manager is
            the only supported source.
        regional_certificate_name(str, Optional):
            The user-friendly name of the certificate that will be used by regional
            endpoint for this domain name.
        regional_certificate_arn(str, Optional):
            The reference to an AWS-managed certificate that will be used by regional
            endpoint for this domain name. AWS Certificate Manager is the only supported source.
        resource_id(str, Optional):
            idem required name for the AWS API Gateway domain
        endpoint_configuration(dict, Optional):
            The endpoint configuration of this DomainName showing the endpoint types of
            the domain name.
            - types(list, Optional):
                A list of endpoint types of an API (RestApi) or its custom domain name
                (DomainName). For an edge-optimized API and its custom domain name, the
                endpoint type is "EDGE" . For a regional API and its custom domain name,
                the endpoint type is REGIONAL . For a private API, the endpoint type
                is PRIVATE .
                string) -- The endpoint type. The valid values are EDGE for edge-optimized
                API setup, most suitable for mobile applications; REGIONAL for regional API
                endpoint setup, most suitable for calling from AWS Region; and PRIVATE
                for private APIs.
            - vpcEndpointIds (list, Optional):
                A list of VpcEndpointIds of an API (RestApi) against which to create Route53
                ALIASes. It is only supported for PRIVATE endpoint type.
        mutual_tls_authentication(dict, Optional):
            The mutual TLS authentication configuration for a custom domain name. If specified,
            API Gateway performs two-way authentication between the client and the server.
            Clients must present a trusted certificate to access your API.
            - TruststoreUri(str, Optional)--
                An Amazon S3 URL that specifies the truststore for mutual TLS authentication,
                for example s3://bucket-name/key-name . The truststore can contain certificates
                from public or private certificate authorities. To update the truststore,
                upload a new version to S3, and then update your custom domain name to use the
                new version. To update the truststore, you must have permissions to access the S3 object.
            - TruststoreVersion(str, Optional):
                The version of the S3 object that contains your truststore. To specify a version,
                you must have versioning enabled for the S3 bucket
        security_policy(str, Optional):
            The Transport Layer Security (TLS) version + cipher suite for this DomainName.
            The valid values are TLS_1_0 and TLS_1_2
        tags(dict, Optional):
            The key-value map of strings. The valid character set is [a-zA-Z+-=._:/]. The tag key can
            be up to 128 characters and must not start with aws: . The tag value can be up to 256 characters.
        Timeout configuration for create/update of AWS ApiGateway Domain Name
                * update(dict, Optional):
                    Timeout configuration for updating AWS ApiGateway Domain Name
                        * delay(int, Optional): The amount of time in seconds to wait between attempts. Default value is ``15``.
                        * max_attempts(int, Optional): Customized timeout configuration containing delay and max attempts. Default value is ``40``.
    Request Syntax:
        .. code-block:: sls

             idem-test.idem-jedi-test.com:
              aws.apigateway.domain_name.present:
              - domain_name: string
              - regional_certificate_name: string
              - regional_certificate_arn: string
              - security_policy: string
              - endpoint_configuration: dict

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

            idem-test.idem-jedi-test.com:
              aws.apigateway.domain_name.present:
              - domain_name: idem-test.idem-jedi-test.com
              - regional_certificate_arn: arn:aws:acm:us-west-2:746014882121:certificate/bb79ca7e-0fd0-485b-a977-cf96ba19f62c
              - security_policy: TLS_1_2
              - endpoint_configuration:
                  types:
                  - REGIONAL
              - tags:
                  -Key: 'idem-resource-name'
                   Value: 'idem-fixture-acm07f548a9-ad28-4b73-bc17-28559c357784.example.com'
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    before_ret = None
    if resource_id:
        before_ret = await hub.exec.aws.apigateway.domain_name.get(
            ctx=ctx, name=name, resource_id=resource_id
        )
        if not before_ret["result"] or not before_ret["ret"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigateway.domain_name", name=name
        )
        result["old_state"] = before_ret["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])
        resource_arn = hub.tool.aws.arn_utils.build(
            service="apigateway",
            region=ctx["acct"]["region_name"],
            resource="/domainnames/" + resource_id,
        )
        update_parameters = {
            "certificate_name": certificate_name,
            "certificate_arn": certificate_arn,
            "endpoint_configuration": endpoint_configuration,
            "mutual_tls_authentication_input": mutual_tls_authentication,
            "regional_certificate_name": regional_certificate_name,
            "regional_certificate_arn": regional_certificate_arn,
            "security_policy": security_policy,
        }
        update_domain_name = await hub.tool.aws.apigateway.domain_name.update(
            ctx,
            name=name,
            old_state=result["old_state"],
            updatable_parameters=update_parameters,
        )
        if not update_domain_name["result"]:
            result["result"] = False
            result["comment"] = update_domain_name["comment"]
            return result
        result["comment"] = result["comment"] + update_domain_name["comment"]
        resource_updated = bool(update_domain_name["ret"])
        if resource_updated and ctx.get("test", False):
            result["new_state"].update(update_domain_name["ret"])
        if tags is not None and tags != result["old_state"].get("tags", {}):
            update_tags_ret = await hub.tool.aws.apigateway.tag.update(
                ctx,
                resource_arn=resource_arn,
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags,
            )
            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] = update_tags_ret["comment"]
                return result
            result["comment"] = result["comment"] + update_tags_ret["comment"]
            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if update_tags_ret["ret"] and ctx.get("test", False):
                result["new_state"]["tags"] = update_tags_ret["ret"]
        if resource_updated:
            if ctx.get("test", False):
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.apigateway.domain_name", name=name
                )
                result["new_state"] = update_parameters
                result["new_state"]["domain_name"] = name
                result["new_state"]["resource_id"] = name
                result["new_state"]["tags"] = tags
                return result
            else:
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.apigateway.domain_name", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "regional_certificate_name": regional_certificate_name,
                    "regional_certificate_arn": regional_certificate_arn,
                    "endpoint_configuration": endpoint_configuration,
                    "certificate_name": certificate_name,
                    "certificate_arn": certificate_arn,
                    "security_policy": security_policy,
                    "tags": tags,
                    "mutual_tls_authentication": mutual_tls_authentication,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.domain_name", name=name
            )
            return result
        else:
            response = await hub.exec.boto3.client.apigateway.create_domain_name(
                ctx,
                domainName=name,
                regionalCertificateName=regional_certificate_name,
                regionalCertificateArn=regional_certificate_arn,
                certificateName=certificate_name,
                certificateArn=certificate_arn,
                endpointConfiguration=endpoint_configuration,
                securityPolicy=security_policy,
                tags=tags,
                mutualTlsAuthentication=mutual_tls_authentication,
            )
            result["result"] = response["result"]
            if not response["result"]:
                result["comment"] = response["comment"]
                return result

            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.apigateway.domain_name", name=name
            )
            resource_id = name
    if (not before_ret) or resource_updated:
        after = await hub.exec.aws.apigateway.domain_name.get(
            ctx, name=name, resource_id=resource_id
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = after["comment"]
            return result
        result["new_state"] = after["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: make_dataclass(
        """Specifies timeout for deletion of domain.""" "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=40)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Deletes an API Gateway domain name resource.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            AWS API Gateway domain name. Idem automatically considers this
            resource being absent if this field is not specified.
        timeout(dict, Optional):
            Timeout configuration for AWS Elasticsearch domain
                * delete(dict, Optional):
                    Timeout configuration when deleting an AWS Elasticsearch domain
                        * delay(int, Optional) -- The amount of time in seconds to wait between attempts. Default value is ``40``.
                        * max_attempts(int, Optional) -- Max attempts of waiting for change. Default value is ``60``.

    Returns:
        Dict[str, Any]

    Request syntax:

        .. code-block:: sls

            idem_test_aws_apigateway_domain_name::
              aws.apigateway.domain_name.absent:
                - name: 'str'
                - resource_id: 'str'

    Examples:

        .. code-block:: sls

            idem-test.idem-jedi-test.com:
              aws.apigateway.domain_name.present:
              - name: idem-test.idem-jedi-test.com
              - resource_id: idem-test.idem-jedi-test.com
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.domain_name", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigateway.domain_name.get(
        ctx=ctx, name=name, resource_id=resource_id
    )
    if not before_ret["result"]:
        result["result"] = False
        result["comment"] = before_ret["comment"]
        return result
    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.domain_name", name=name
        )

    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.domain_name", name=name
        )
        return result
    else:
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigateway.delete_domain_name(
            ctx, domainName=resource_id
        )
        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.domain_name", name=name
        )
        delete_waiter_acceptors = [
            {
                "matcher": "path",
                "expected": "AVAILABLE",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "UPDATING",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "PENDING",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "PENDING_CERTIFICATE_REIMPORT",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "path",
                "expected": "PENDING_OWNERSHIP_VERIFICATION",
                "state": "retry",
                "argument": "domainNameStatus",
            },
            {
                "matcher": "error",
                "expected": "NotFoundException",
                "state": "success",
                "argument": "Error.Code",
            },
            {
                "matcher": "error",
                "expected": "BadRequestException",
                "state": "retry",
                "argument": "Error.Code",
            },
            {
                "matcher": "error",
                "expected": "UnauthorizedException",
                "state": "retry",
                "argument": "Error.Code",
            },
            {
                "matcher": "error",
                "expected": "TooManyRequestsException",
                "state": "retry",
                "argument": "Error.Code",
            },
        ]

        delete_waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=40,
            default_max_attempts=60,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        delete_domain_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="ApiGatewayDomainNameDelete",
            operation="GetDomainName",
            argument=["domainNameStatus"],
            acceptors=delete_waiter_acceptors,
            client=await hub.tool.boto3.client.get_client(ctx, "apigateway"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "apigateway",
                "ApiGatewayDomainNameDelete",
                delete_domain_waiter,
                domainName=resource_id,
                WaiterConfig=delete_waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe all the API Gateway domain_name.

    Returns:
        Dict[str, Dict[str, Any]]

    Request Syntax:

        .. code-block:: sls

            $ idem describe aws.apigateway.domain_name
    """

    result = {}
    get_domain_names_ret = await hub.exec.aws.apigateway.domain_name.list(ctx)
    for domain_name in get_domain_names_ret["ret"]:
        domain_name["resource_id"]
        result[domain_name["resource_id"]] = {
            "aws.apigateway.domain_name.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in domain_name.items()
            ]
        }
    return result



"""States module for managing Rds Db Proxys."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_proxy_name: str,
    engine_family: str,
    auth: List[
        make_dataclass(
            "UserAuthConfig",
            [
                ("Description", str, field(default=None)),
                ("UserName", str, field(default=None)),
                ("AuthScheme", str, field(default=None)),
                ("SecretArn", str, field(default=None)),
                ("IAMAuth", str, field(default=None)),
                ("ClientPasswordAuthType", str, field(default=None)),
            ],
        )
    ],
    role_arn: str,
    vpc_subnet_ids: List[str],
    vpc_security_group_ids: List[str] = None,
    require_tls: bool = None,
    idle_client_timeout: int = None,
    debug_logging: bool = None,
    tags: Dict[str, str] = None,
    resource_id: str = None,
    timeout: make_dataclass(
        """Timeout configuration.""" "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=None)),
                        ("max_attempts", int, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """
    Creates a new DB proxy.

    Args:
        db_proxy_name(str): The identifier for the proxy. This name must be unique for all proxies owned by your Amazon Web
            Services account in the specified Amazon Web Services Region. An identifier must begin with a
            letter and must contain only ASCII letters, digits, and hyphens; it can't end with a hyphen or
            contain two consecutive hyphens.

        engine_family(str): The kinds of databases that the proxy can connect to. This value determines which database
            network protocol the proxy recognizes when it interprets network traffic to and from the
            database. For Aurora MySQL, RDS for MariaDB, and RDS for MySQL databases, specify MYSQL. For
            Aurora PostgreSQL and RDS for PostgreSQL databases, specify POSTGRESQL. For RDS for Microsoft
            SQL Server, specify SQLSERVER.

        auth(List[dict[str, Any]]): The authorization mechanism that the proxy uses.

            * Description (str, Optional): A user-specified description about the authentication used by a proxy to log in as a specific
            database user.

            * UserName (str, Optional): The name of the database user to which the proxy connects.

            * AuthScheme (str, Optional): The type of authentication that the proxy uses for connections from the proxy to the underlying
            database.

            * SecretArn (str, Optional): The Amazon Resource Name (ARN) representing the secret that the proxy uses to authenticate to
            the RDS DB instance or Aurora DB cluster. These secrets are stored within Amazon Secrets
            Manager.

            * IAMAuth (str, Optional): Whether to require or disallow Amazon Web Services Identity and Access Management (IAM)
            authentication for connections to the proxy. The ENABLED value is valid only for proxies with
            RDS for Microsoft SQL Server.

            * ClientPasswordAuthType (str, Optional): The type of authentication the proxy uses for connections from clients.

        role_arn(str): The Amazon Resource Name (ARN) of the IAM role that the proxy uses to access secrets in Amazon
            Web Services Secrets Manager.

        vpc_subnet_ids(List[str]): One or more VPC subnet IDs to associate with the new proxy.

        name(str): Idem name of the resource.

        vpc_security_group_ids(List[str], Optional): One or more VPC security group IDs to associate with the new proxy. Defaults to None.

        require_tls(bool, Optional): A Boolean parameter that specifies whether Transport Layer Security (TLS) encryption is required
            for connections to the proxy. By enabling this setting, you can enforce encrypted TLS
            connections to the proxy. Defaults to None.

        idle_client_timeout(int, Optional): The number of seconds that a connection to the proxy can be inactive before the proxy
            disconnects it. You can set this value higher or lower than the connection timeout limit for the
            associated database. Defaults to None.

        debug_logging(bool, Optional): Whether the proxy includes detailed information about SQL statements in its logs. This
            information helps you to debug issues involving SQL behavior or the performance and scalability
            of the proxy connections. The debug information includes the text of SQL statements that you
            submit through the proxy. Thus, only enable this setting when needed for debugging, and only
            when you have security measures in place to safeguard any sensitive information that appears in
            the logs. Defaults to None.

        tags(Dict[str, str], Optional): The tags to apply to the resource.

        resource_id(str, Optional): Db_proxy unique ID. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for create/update of AWS DB Cluster.

            * create (*dict, Optional*):
                Timeout configuration for creating DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

            * update(*dict, Optional*):
                Timeout configuration for updating DB Cluster

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls


          idem_test_aws.rds.db_proxy_is_present:
              aws.rds.db_proxy.present:
              - db_proxy_name: string
              - engine_family: string
              - auth:
                - auth_scheme: string
                  client_password_auth_type: string
                  description: string
                  iam_auth: string
                  secret_arn: string
                  user_name: string
              - role_arn: string
              - vpc_subnet_ids:
                - value
              - vpc_security_group_ids:
                - value
              - require_tls: bool
              - idle_client_timeout: int
              - debug_logging: bool
              - tags:
                - key: string
                  value: string


    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    current_state = None

    if resource_id:
        before = await hub.exec.aws.rds.db_proxy.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = current_state = copy.deepcopy(before["ret"])
        result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.rds.db_proxy", name=name
        )

    if current_state:
        # If there are changes in desired state from existing state
        changes = differ.deep_diff(
            current_state if current_state else {}, desired_state
        )

        if bool(changes.get("new")):
            if ctx.test:
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={}, desired_state=desired_state
                )
                result["comment"] += hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.rds.db_proxy", name=name
                )
                return result
            else:
                # Update the resource
                update_ret = await hub.exec.aws.rds.db_proxy.update(
                    ctx,
                    resource_id=resource_id,
                    new_db_proxy_name=db_proxy_name,
                    auth=auth,
                    require_tls=require_tls,
                    idle_client_timeout=idle_client_timeout,
                    debug_logging=debug_logging,
                    role_arn=role_arn,
                    security_groups=vpc_security_group_ids,
                    name=name,
                    tags=tags,
                )
                result["result"] = update_ret["result"]

                if result["result"]:
                    resource_id = update_ret["ret"].get("resource_id", resource_id)
                    result["comment"] += hub.tool.aws.comment_utils.update_comment(
                        resource_type="aws.rds.db_proxy", name=name
                    )
                else:
                    result["comment"] += update_ret["comment"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.rds.db_proxy", name=name
            )
            return result

        create_ret = await hub.exec.aws.rds.db_proxy.create(
            ctx, **desired_state, timeout=timeout
        )
        result["result"] = create_ret["result"]

        if result["result"]:
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.rds.db_proxy", name=name
            )

            resource_id = create_ret["ret"].get("resource_id")
            # Safeguard for any future errors so that the resource_id is saved in the ESM
            result["new_state"] = dict(name=name, resource_id=resource_id)
        else:
            result["comment"].append(create_ret["comment"])

    if not result["result"]:
        # If there is any failure in create/update, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = dict(name=name, resource_id=resource_id)

    after = await hub.exec.aws.rds.db_proxy.get(ctx, name=name, resource_id=resource_id)
    result["new_state"] = after.ret
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: make_dataclass(
        """Specifies timeout for deletion of DB Proxy.""" "Timeout",
        [
            (
                "delete",
                make_dataclass(
                    "DeleteTimeout",
                    [
                        ("delay", int, field(default=40)),
                        ("max_attempts", int, field(default=60)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """

    Deletes an existing DB proxy.

    Args:
        name(str): Idem name of the resource.

        resource_id(str, Optional): Db_proxy unique ID. Defaults to None.

        timeout(dict, Optional):
            Timeout configuration for delete of AWS RDS DB Proxy.

            * delete (*dict, Optional*):
                Timeout configuration for deleting DB Cluster.

                * delay(*int, Optional*):
                    The amount of time in seconds to wait between attempts.

                * max_attempts(*int, Optional*):
                    Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

            idem_test_aws.rds.db_proxy_is_absent:
              aws.rds.db_proxy.absent:
              - db_proxy_name: string


    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # This is to make absent idempotent. If absent is run again, it would be a no-op
    if not resource_id:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_proxy", name=name
        )
        return result

    before = await hub.exec.aws.rds.db_proxy.get(
        ctx, name=name, resource_id=resource_id
    )

    # Case: Error
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    # Case: Not Found
    if not before["ret"]:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_proxy", name=name
        )
        return result

    result["old_state"] = before["ret"]
    if ctx.get("test", False):
        result["comment"] += hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.rds.db_proxy", name=name
        )
        return result

    delete_ret = await hub.exec.aws.rds.db_proxy.delete(
        ctx, name=name, resource_id=resource_id, timeout=timeout
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        # If there is any failure in delete, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = resource_id
        result["comment"].append(delete_ret["comment"])
    else:
        result["comment"] += hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.rds.db_proxy", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns information about DB proxies.

    Returns:
        Dict[str, Any]

    Example:

        .. code-block:: bash

            $ idem describe aws.rds.db_proxy
    """

    result = {}
    ret = await hub.exec.aws.rds.db_proxy.list(ctx)

    if not ret or not ret["result"]:
        hub.log.debug(f"Could not describe aws.rds.db_proxy {ret['comment']}")
        return result

    for resource in ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "aws.rds.db_proxy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""State module for managing AWS S3 bucket website."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from dict_tools import data

__contracts__ = ["resource"]

STATE_NAME = "aws.s3.bucket_website"


async def present(
    hub,
    ctx,
    name: str,
    bucket: str,
    website_configuration: make_dataclass(
        "WebsiteConfiguration",
        [
            (
                "ErrorDocument",
                make_dataclass("ErrorDocument", [("Key", str)]),
                field(default=None),
            ),
            (
                "IndexDocument",
                make_dataclass("IndexDocument", [("Suffix", str)]),
                field(default=None),
            ),
            (
                "RedirectAllRequestsTo",
                make_dataclass(
                    "RedirectAllRequestsTo",
                    [("HostName", str), ("Protocol", str, field(default=None))],
                ),
                field(default=None),
            ),
            (
                "RoutingRules",
                List[
                    make_dataclass(
                        "RoutingRule",
                        [
                            (
                                "Redirect",
                                make_dataclass(
                                    "Redirect",
                                    [
                                        ("HostName", str, field(default=None)),
                                        ("HttpRedirectCode", str, field(default=None)),
                                        ("Protocol", str, field(default=None)),
                                        (
                                            "ReplaceKeyPrefixWith",
                                            str,
                                            field(default=None),
                                        ),
                                        ("ReplaceKeyWith", str, field(default=None)),
                                    ],
                                ),
                            ),
                            (
                                "Condition",
                                make_dataclass(
                                    "Condition",
                                    [
                                        (
                                            "HttpErrorCodeReturnedEquals",
                                            str,
                                            field(default=None),
                                        ),
                                        ("KeyPrefixEquals", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                        ],
                    )
                ],
                field(default=None),
            ),
        ],
    ),
    resource_id: str = None,
    checksum_algorithm: str = None,
    expected_bucket_owner: str = None,
) -> Dict[str, Any]:
    """Sets the configuration of the website that is specified in the website subresource. To configure a bucket as a
    website, you can add this subresource on the bucket with website configuration information such as the file name
    of the index document and any redirect rules. For more information, see Hosting Websites on Amazon S3. This PUT
    action requires the S3:PutBucketWebsite permission. By default, only the bucket owner can configure the website
    attached to a bucket; however, bucket owners can allow other users to set the website configuration by writing a
    bucket policy that grants them the S3:PutBucketWebsite permission. To redirect all website requests sent to the
    bucket's website endpoint, you add a website configuration with the following elements. Because all requests are
    sent to another website, you don't need to provide index document name for the bucket.    WebsiteConfiguration
    RedirectAllRequestsTo     HostName     Protocol    If you want granular control over redirects, you can use the
    following elements to add routing rules that describe conditions for redirecting requests and information about
    the redirect destination. In this case, the website configuration must provide an index document for the bucket,
    because some requests might not be redirected.     WebsiteConfiguration     IndexDocument     Suffix
    ErrorDocument     Key     RoutingRules     RoutingRule     Condition     HttpErrorCodeReturnedEquals
    KeyPrefixEquals     Redirect     Protocol     HostName     ReplaceKeyPrefixWith     ReplaceKeyWith
    HttpRedirectCode    Amazon S3 has a limitation of 50 routing rules per website configuration. If you require
    more than 50 routing rules, you can use object redirect. For more information, see Configuring an Object
    Redirect in the Amazon S3 User Guide.

        Args:
            name(str):
                An Idem name of the resource.

            bucket(str):
                The bucket name.

            resource_id(str, Optional):
                The bucket name.

            checksum_algorithm(str, Optional):
                Indicates the algorithm used to create the checksum for the object when using the SDK. This
                header will not provide any additional functionality if not using the SDK. When sending this
                header, there must be a corresponding x-amz-checksum or x-amz-trailer header sent. Otherwise,
                Amazon S3 fails the request with the HTTP status code 400 Bad Request. For more information, see
                Checking object integrity in the Amazon S3 User Guide. If you provide an individual checksum,
                Amazon S3 ignores any provided ChecksumAlgorithm parameter. Defaults to None.

            website_configuration(Dict[str, Any]):
                Container for the request.

                * ErrorDocument (dict[str, any], Optional):
                    The name of the error document for the website.

                    * Key (str):
                        The object key name to use when a 4XX class error occurs.  Replacement must be made for object
                keys containing special characters (such as carriage returns) when using XML requests. For more
                information, see  XML related object key constraints.

                * IndexDocument (dict[str, any], Optional):
                    The name of the index document for the website.

                    * Suffix (str):
                        A suffix that is appended to a request that is for a directory on the website endpoint (for
                example,if the suffix is index.html and you make a request to samplebucket/images/ the data that
                is returned will be for the object with the key name images/index.html) The suffix must not be
                empty and must not include a slash character.  Replacement must be made for object keys
                containing special characters (such as carriage returns) when using XML requests. For more
                information, see  XML related object key constraints.

                * RedirectAllRequestsTo (dict[str, any], Optional):
                    The redirect behavior for every request to this bucket's website endpoint.  If you specify this
                property, you can't specify any other property.

                    * HostName (str):
                        Name of the host where requests are redirected.

                    * Protocol (str, Optional):
                        Protocol to use when redirecting requests. The default is the protocol that is used in the
                original request.

                * RoutingRules (list[dict[str, any]], Optional):
                    Rules that define when a redirect is applied and the redirect behavior.

                    * Condition (dict[str, any], Optional):
                        A container for describing a condition that must be met for the specified redirect to apply. For
                example, 1. If request is for pages in the /docs folder, redirect to the /documents folder. 2.
                If request results in HTTP error 4xx, redirect request to another host where you might process
                the error.

                        * HttpErrorCodeReturnedEquals (str, Optional):
                            The HTTP error code when the redirect is applied. In the event of an error, if the error code
                equals this value, then the specified redirect is applied. Required when parent element
                Condition is specified and sibling KeyPrefixEquals is not specified. If both are specified, then
                both must be true for the redirect to be applied.

                        * KeyPrefixEquals (str, Optional):
                            The object key name prefix when the redirect is applied. For example, to redirect requests for
                ExamplePage.html, the key prefix will be ExamplePage.html. To redirect request for all pages
                with the prefix docs/, the key prefix will be /docs, which identifies all objects in the docs/
                folder. Required when the parent element Condition is specified and sibling
                HttpErrorCodeReturnedEquals is not specified. If both conditions are specified, both must be
                true for the redirect to be applied.  Replacement must be made for object keys containing
                special characters (such as carriage returns) when using XML requests. For more information, see
                XML related object key constraints.

                    * Redirect (dict[str, any]):
                        Container for redirect information. You can redirect requests to another host, to another page,
                or with another protocol. In the event of an error, you can specify a different error code to
                return.

                        * HostName (str, Optional):
                            The host name to use in the redirect request.

                        * HttpRedirectCode (str, Optional):
                            The HTTP redirect code to use on the response. Not required if one of the siblings is present.

                        * Protocol (str, Optional):
                            Protocol to use when redirecting requests. The default is the protocol that is used in the
                original request.

                        * ReplaceKeyPrefixWith (str, Optional):
                            The object key prefix to use in the redirect request. For example, to redirect requests for all
                pages with prefix docs/ (objects in the docs/ folder) to documents/, you can set a condition
                block with KeyPrefixEquals set to docs/ and in the Redirect set ReplaceKeyPrefixWith to
                /documents. Not required if one of the siblings is present. Can be present only if
                ReplaceKeyWith is not provided.  Replacement must be made for object keys containing special
                characters (such as carriage returns) when using XML requests. For more information, see  XML
                related object key constraints.

                        * ReplaceKeyWith (str, Optional):
                            The specific object key to use in the redirect request. For example, redirect request to
                error.html. Not required if one of the siblings is present. Can be present only if
                ReplaceKeyPrefixWith is not provided.  Replacement must be made for object keys containing
                special characters (such as carriage returns) when using XML requests. For more information, see
                XML related object key constraints.

            expected_bucket_owner(str, Optional):
                The account ID of the expected bucket owner. If the bucket is owned by a different account, the
                request fails with the HTTP status code 403 Forbidden (access denied). Defaults to None.

        Request Syntax:
            .. code-block:: sls

                resource_is_present:
                  aws.s3.bucket_website.present:
                    - name: "str"
                    - resource_id: "str"
                    - bucket: "string"
                    - checksum_algorithm: "string"
                    - website_configuration: "Dict"
                    - expected_bucket_owner: "string"

        Returns:
            Dict[str, Any]


        Examples:
            .. code-block:: sls

                resource_is_present:
                  aws.s3.bucket_website.present:
                    - name: value
                    - bucket: value
                    - website_configuration:
                        ErrorDocument:
                            Key: "index.html"
                        IndexDocument:
                            Suffix: "index.html"
                        RedirectAllRequestsTo:
                            HostName: 'test.com'
                            Protocol: https
                        RoutingRules:
                            Condition:
                                HttpErrorCodeReturnedEquals: '400'
                                KeyPrefixEquals: 'test'
                            Redirect:
                                HostName: 'test.example.com'
                                HttpRedirectCode: 500
                                Protocol: https
                                ReplaceKeyPrefixWith: 'test1'
                                ReplaceKeyWith: 'test'


    """
    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        if resource_id != bucket:
            result["result"] = False
            result["comment"].append(
                f"Bucket '{bucket}' and resource_id '{resource_id}' parameters must be the same",
            )
            return result

        before = await hub.exec.aws.s3.bucket_website.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

        result["old_state"] = copy.deepcopy(before["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        result["comment"] += list(
            hub.tool.aws.comment_utils.already_exists_comment("aws.s3.bucket", name)
        )
        if not data.recursive_diff(
            website_configuration,
            result["old_state"]["website_configuration"],
            ignore_missing_keys=True,
            ignore_order=True,
        ):
            # no updates to make. return from here
            return result

        if ctx.get("test", False):
            result["new_state"]["website_configuration"] = website_configuration
            result["comment"] += list(
                hub.tool.aws.comment_utils.would_update_comment(
                    resource_type=STATE_NAME, name=name
                )
            )
            return result
        else:
            result["comment"] += list(
                hub.tool.aws.comment_utils.update_comment(
                    resource_type=STATE_NAME, name=name
                )
            )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "bucket": bucket,
                    "website_configuration": website_configuration,
                    "checksum_algorithm": checksum_algorithm,
                    "expected_bucket_owner": expected_bucket_owner,
                },
            )
            result["comment"] = list(
                hub.tool.aws.comment_utils.would_create_comment(
                    resource_type=STATE_NAME, name=name
                )
            )
            return result

        result["comment"] = list(
            hub.tool.aws.comment_utils.create_comment(
                resource_type=STATE_NAME, name=name
            )
        )

    put_ret = await hub.exec.boto3.client.s3.put_bucket_website(
        ctx,
        Bucket=bucket,
        ChecksumAlgorithm=checksum_algorithm,
        WebsiteConfiguration=website_configuration,
        ExpectedBucketOwner=expected_bucket_owner,
    )
    if not put_ret["result"]:
        result["result"] = False
        result["comment"] += list(put_ret["comment"])
        return result
    resource_id = bucket
    after = await hub.exec.aws.s3.bucket_website.get(
        ctx, name=name, resource_id=resource_id
    )
    if not after["result"] or not after["ret"]:
        result["comment"] += after["comment"]
        result["result"] = False
        return result

    result["new_state"] = after["ret"]

    return result


async def absent(
    hub, ctx, name: str, expected_bucket_owner: str = None, resource_id: str = None
) -> Dict[str, Any]:
    """This action removes the website configuration for a bucket. Amazon S3 returns a 200 OK response upon
    successfully deleting a website configuration on the specified bucket. You will get a 200 OK response if the
    website configuration you are trying to delete does not exist on the bucket. Amazon S3 returns a 404 response if
    the bucket specified in the request does not exist. This DELETE action requires the S3:DeleteBucketWebsite
    permission. By default, only the bucket owner can delete the website configuration attached to a bucket.
    However, bucket owners can grant other users permission to delete the website configuration by writing a bucket
    policy granting them the S3:DeleteBucketWebsite permission.  For more information about hosting websites, see
    Hosting Websites on Amazon S3.  The following operations are related to DeleteBucketWebsite:    GetBucketWebsite
    PutBucketWebsite

        Args:
            name(str):
                An Idem name of the resource.

            expected_bucket_owner(str, Optional):
                The account ID of the expected bucket owner. If the bucket is owned by a different account, the
                request fails with the HTTP status code 403 Forbidden (access denied). Defaults to None.

            resource_id(str, Optional):
                The bucket name for which you want to remove the website configuration.


        Request Syntax:
            .. code-block:: sls

                resource_is_absent:
                  aws.s3.bucket_website.absent:
                    - name: "str"
                    - resource_id: "str"

        Returns:
            Dict[str, Any]

        Examples:
            .. code-block:: sls

                resource_is_absent:
                  aws.s3.bucket_website.absent:
                    - name: value
                    - resource_id: value
    """

    result = dict(comment=[], old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type=STATE_NAME, name=name
            )
        )
        return result

    before = await hub.exec.aws.s3.bucket_website.get(
        ctx, name=name, resource_id=resource_id
    )
    if not before["result"]:
        if "NoSuchBucket" in str(before["comment"]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type=STATE_NAME, name=name
            )
            return result
        result["result"] = False
        result["comment"] += before["comment"]
        return result

    if not before["ret"]:
        result["comment"] += list(
            hub.tool.aws.comment_utils.already_absent_comment(
                resource_type=STATE_NAME, name=name
            )
        )
    elif ctx.get("test", False):
        result["old_state"] = before["ret"]
        result["comment"] += list(
            hub.tool.aws.comment_utils.would_delete_comment(
                resource_type=STATE_NAME, name=name
            )
        )
        return result
    else:
        result["old_state"] = before["ret"]
        ret = await hub.exec.boto3.client.s3.delete_bucket_website(
            ctx, Bucket=resource_id, ExpectedBucketOwner=expected_bucket_owner
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] += list(ret["comment"])
            return result
        result["comment"] += list(
            hub.tool.aws.comment_utils.delete_comment(
                resource_type=STATE_NAME, name=name
            )
        )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Gets the website configuration for each S3 bucket under the given AWS account.

    Returns:
        dict[str, dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.s3.bucket_website
    """
    result = {}

    list_buckets_ret = await hub.exec.boto3.client.s3.list_buckets(ctx)
    if not list_buckets_ret["result"]:
        hub.log.debug(f"Could not list S3 buckets: {list_buckets_ret['comment']}")
        return result

    for bucket in list_buckets_ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")

        get_bucket_website_ret = await hub.exec.aws.s3.bucket_website.get(
            ctx, name=bucket_name, resource_id=bucket_name
        )
        if not get_bucket_website_ret["result"] or not get_bucket_website_ret["ret"]:
            hub.log.debug(
                f"Could not get website configuration for S3 bucket '{bucket_name}': "
                f"{get_bucket_website_ret['comment']}. Describe will skip this S3 bucket and continue."
            )
            continue

        resource_translated = get_bucket_website_ret["ret"]

        result[resource_translated["name"]] = {
            f"{STATE_NAME}.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



"""States module for managing Rds Db Proxy Targets."""
import copy
from typing import Any
from typing import Dict
from typing import List

from dict_tools import differ

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    db_proxy_name: str,
    target_group_name: str,
    db_instance_identifiers: List[str] = None,
    db_cluster_identifiers: List[str] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Associate one or more DBProxyTarget data structures with a DBProxyTargetGroup.

    Args:
        name(str): Idem name of the resource.

        db_proxy_name(str): The identifier of the DBProxy that is associated with the DBProxyTargetGroup.

        target_group_name(str, Optional): The identifier of the DBProxyTargetGroup. Defaults to None.

        db_instance_identifiers(List[str], Optional): One or more DB instance identifiers. Defaults to None.

        db_cluster_identifiers(List[str], Optional): One or more DB cluster identifiers. Defaults to None.

        resource_id(str, Optional): Db_proxy_target unique ID. Defaults to None.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls


          idem_test_aws_auto.rds.db_proxy_target_is_present:
              aws_auto.aws_auto.rds.db_proxy_target.present:
              - db_proxy: string
              - target_group_name: string
              - db_instance_identifiers:
                - value
              - db_cluster_identifiers:
                - value
              - tags: dict


    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    current_state = None

    if resource_id:
        before = await hub.exec.aws.rds.db_proxy_target.get(
            ctx, name=name, resource_id=resource_id
        )
        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = current_state = copy.deepcopy(before["ret"])
        result["comment"] += hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.rds.db_proxy_target", name=name
        )
        result["new_state"] = copy.deepcopy(result["old_state"])
        return result

    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={}, desired_state=desired_state
        )
        result["comment"] += hub.tool.aws.comment_utils.would_create_comment(
            resource_type="aws.rds.db_proxy_target", name=name
        )
        return result

    create_ret = await hub.exec.aws.rds.db_proxy_target.create(
        ctx,
        **desired_state,
    )
    result["result"] = create_ret["result"]

    if result["result"]:
        result["comment"] += hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.rds.db_proxy_target", name=name
        )
        resource_id = create_ret["ret"].get("resource_id")
        # Safeguard for any future errors so that the resource_id is saved in the ESM
        result["new_state"] = dict(name=name, resource_id=resource_id)
    else:
        result["comment"].append(create_ret["comment"])

    if not result["result"]:
        # If there is any failure in create/update, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = dict(name=name, resource_id=resource_id)

    after = await hub.exec.aws.rds.db_proxy_target.get(
        ctx, name=name, resource_id=resource_id
    )
    result["new_state"] = after.ret
    return result


async def absent(
    hub,
    ctx,
    name: str,
    db_instance_identifiers: List[str] = None,
    db_cluster_identifiers: List[str] = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """

    Remove the association between one or more DBProxyTarget data structures and a DBProxyTargetGroup.

    Args:
        name(str): Idem name of the resource.

        db_instance_identifiers(List[str], Optional): One or more DB instance identifiers. Defaults to None.

        db_cluster_identifiers(List[str], Optional): One or more DB cluster identifiers. Defaults to None.

        resource_id(str, Optional): Db_proxy_target unique ID. Defaults to None.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls


            idem_test_aws_auto.rds.db_proxy_target_is_absent:
              aws_auto.aws_auto.rds.db_proxy_target.absent:
              - db_proxy_name: string
              - target_group_name: string
              - db_instance_identifiers:
                - value
              - db_cluster_identifiers:
                - value


    """

    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    # This is to make absent idempotent. If absent is run again, it would be a no-op
    if not resource_id:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_proxy_target", name=name
        )
        return result

    before = await hub.exec.aws.rds.db_proxy_target.get(
        ctx, name=name, resource_id=resource_id
    )

    # Case: Error
    if not before["result"]:
        result["result"] = False
        result["comment"] = before["comment"]
        return result

    # Case: Not Found
    if not before["ret"]:
        result["comment"] += hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.rds.db_proxy_target", name=name
        )
        return result

    result["old_state"] = before["ret"]

    if ctx.get("test", False):
        result["comment"].append(
            f"Would delete aws.rds.db_proxy_target '{name}'",
        )
        return result

    delete_ret = await hub.exec.aws.rds.db_proxy_target.delete(
        ctx,
        name=name,
        resource_id=resource_id,
        db_cluster_identifiers=db_cluster_identifiers,
        db_instance_identifiers=db_instance_identifiers,
    )

    result["result"] = delete_ret["result"]

    if not result["result"]:
        # If there is any failure in delete, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = resource_id
        result["comment"].append(delete_ret["comment"])
    else:
        result["comment"] += hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.rds.db_proxy_target", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Returns information about DBProxyTarget objects. This API supports pagination.

    Returns:
        Dict[str, Any]

    Example:

        .. code-block:: bash

            $ idem describe aws_auto.rds.db_proxy_target
    """

    result = {}

    db_proxy_ret = await hub.exec.aws.rds.db_proxy.list(ctx)

    if not db_proxy_ret or not db_proxy_ret["result"]:
        hub.log.debug(f"Could not describe aws.rds.db_proxy {db_proxy_ret['comment']}")
        return result
    for db_proxy in db_proxy_ret["ret"]:
        ret = await hub.exec.aws.rds.db_proxy_target.list(
            ctx, db_proxy_name=db_proxy["resource_id"]
        )
        if not ret or not ret["result"]:
            hub.log.debug(
                f"Could not describe aws.rds.db_proxy_target {ret['comment']}"
            )
            return result
        for resource in ret["ret"]:
            resource_id = resource.get("resource_id")
            result[resource_id] = {
                "aws.rds.db_proxy_target.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource.items()
                ]
            }
    return result



"""Util module to provide support to AWS WAFV2 IPSet state module."""
from collections import OrderedDict
from typing import Any
from typing import Dict


async def update(
    hub,
    ctx,
    name: str,
    current_state: Dict[str, Any],
    desired_state: Dict[str, Any],
):
    """Updates configuration options for an ip_set in AWS.

    Args:
        hub:
             The redistributed pop central hub.
        ctx:
             A dict with the keys/values for the execution of the Idem run located in
              `hub.idem.RUNS[ctx['run_name']]`.
        name(str):
             The name of the ip_set in AWS
        current_state(Dict):
             Previous state of the ip_set resource in AWS
        desired_state(Dict):
             Parameters from SLS file as per the desired final state of the ip_set.

    Returns:
        Dict[str, Any]
    """
    result = dict(comment=(), result=True, ret=None)

    parameters = OrderedDict(
        {
            "addresses": "Addresses",
            "description": "Description",
        }
    )

    parameters_to_update = {}

    if desired_state.get("addresses"):
        desired_set_addresses = set(desired_state.get("addresses"))
        current_set_addresses = set(current_state.get("addresses"))
        if (
            desired_set_addresses is not None
            and desired_set_addresses != current_set_addresses
        ):
            parameters_to_update[parameters["addresses"]] = desired_state.get(
                "addresses"
            )
    if desired_state.get("description"):
        if desired_state.get("description") is not None and desired_state.get(
            "description"
        ) != current_state.get("description"):
            parameters_to_update[parameters["description"]] = desired_state.get(
                "description"
            )

    if parameters_to_update:
        if ctx.get("test", False):
            result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
                resource_type="aws.wafv2.ip_set",
                name=name,
            )
            result["ret"] = parameters_to_update
        if not ctx.get("test", False):
            update_ret = await hub.exec.boto3.client.wafv2.update_ip_set(
                ctx=ctx,
                Name=current_state["name"],
                Scope=current_state["scope"],
                Id=current_state["resource_id"],
                LockToken=current_state["lock_token"],
                **parameters_to_update,
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] = update_ret["comment"]
                return result
            result["comment"] = hub.tool.aws.comment_utils.update_comment(
                resource_type="aws.wafv2.ip_set",
                name=name,
            )
            result["ret"] = parameters_to_update
    else:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.wafv2.ip_set",
            name=name,
        )
    return result



from typing import Any
from typing import Dict
from typing import List

# Update all these attributes together
__update_group__ = [
    "affinity",
    "host_id",
    "tenancy",
    "group_name",
    "partition_number",
    "host_resource_group_arn",
]
__update_state__ = "stopped"


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: Dict[str, Any],
    new_value: Dict[str, Any],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    result = True

    group_name = new_value.get("group_name")
    if "group_name" in old_value and not new_value.get("group_name"):
        # An empty string removes the instance from the placement group
        group_name = ""

    # Modify the placement
    ret = await hub.exec.boto3.client.ec2.modify_instance_placement(
        ctx,
        InstanceId=resource.id,
        Affinity=new_value.get("affinity"),
        GroupName=group_name,
        HostId=new_value.get("host_id"),
        Tenancy=new_value.get("tenancy"),
        PartitionNumber=new_value.get("partition_number"),
        HostResourceGroupArn=new_value.get("host_resource_group_arn"),
    )
    if ret.comment:
        comments.append(ret.comment)
    result &= ret.result

    return result



"""State module for managing Organization Configuration."""
from dataclasses import field
from typing import Any
from typing import Dict

__contracts__ = ["resource"]

from dataclasses import make_dataclass


async def present(
    hub,
    ctx,
    name: str,
    auto_enable: bool,
    resource_id: str = None,
    data_sources: make_dataclass(
        """Describes which data sources will be enabled for the detector."""
        "DataSourceConfiguration",
        [
            (
                "S3Logs",
                make_dataclass(
                    """Describes whether S3 data event logs are enabled as a data source."""
                    "S3LogsConfiguration",
                    [("Enable", bool)],
                ),
                field(default=None),
            ),
            (
                "Kubernetes",
                make_dataclass(
                    """Describes whether any Kubernetes logs are enabled as data sources."""
                    "KubernetesConfiguration",
                    [
                        (
                            "AuditLogs",
                            make_dataclass(
                                """The status of Kubernetes audit logs as a data source."""
                                "KubernetesAuditLogsConfiguration",
                                [("Enable", bool)],
                            ),
                        )
                    ],
                ),
                field(default=None),
            ),
            (
                "MalwareProtection",
                make_dataclass(
                    """Describes whether Malware Protection is enabled as a data source."""
                    "MalwareProtectionConfiguration",
                    [
                        (
                            "ScanEc2InstanceWithFindings",
                            make_dataclass(
                                """Describes the configuration of Malware Protection for EC2 instances with findings."""
                                "ScanEc2InstanceWithFindingsConfiguration",
                                [("EbsVolumes", bool, field(default=None))],
                            ),
                            field(default=None),
                        )
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    """Updates the delegated administrator account with the values provided.

    Args:
        name(str):
            An Idem name of the resource.
        auto_enable(bool):
            Indicates whether to automatically enable member accounts in the organization.
        resource_id(str, Optional):
            The ID of the detector to update the delegated administrator for.
        data_sources(dict, Optional):
            Describes which data sources will be updated.

            * S3Logs (*dict, Optional*):
                Describes whether S3 data event logs are enabled as a data source.

                * Enable (*bool*): The status of S3 data event logs as a data source.

            * Kubernetes (*dict, Optional*):
                Describes whether any Kubernetes logs are enabled as data sources.

                * AuditLogs (*dict*):
                    The status of Kubernetes audit logs as a data source.

                    * Enable (*bool*):
                        The status of Kubernetes audit logs as a data source.

            * MalwareProtection (*dict, Optional*):
                Describes whether Malware Protection is enabled as a data source.

                * ScanEc2InstanceWithFindings (*dict, Optional*):
                    Describes the configuration of Malware Protection for EC2 instances with findings.

                    EbsVolumes (*bool, Optional*):
                        Describes the configuration for scanning EBS volumes as data source.

    Request Syntax:

        Using in a state:

        .. code-block:: yaml

             aws.guardduty.organization_configuration.present:
                - name: 'string'
                - resource_id: 'string'
                - auto_enable: True|False
                - data_sources:
                    S3Logs:
                        Enable: True|False
                    Kubernetes:
                        AuditLogs:
                            Enable: True|False
                    MalwareProtection:
                        ScanEc2InstanceWithFindings:
                            EbsVolumes: True|False

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.guardduty.organization_configuration.present:
                - name: 6ec3816a325af96978c683c9c81fdf0a
                - resource_id: 6ec3816a325af96978c683c9c81fdf0a
                - auto_enable: false
                - data_sources:
                    Kubernetes:
                        AuditLogs:
                            AutoEnable: false
                    MalwareProtection:
                        ScanEc2InstanceWithFindings:
                            EbsVolumes:
                                AutoEnable: false
                    S3Logs:
                        AutoEnable: false
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        resource_id = name

    before = await hub.exec.aws.guardduty.organization_configuration.get(
        ctx, resource_id=resource_id
    )
    if not before["result"]:
        result["result"] = before["result"]
        result["comment"] = before["comment"]
        return result

    result["old_state"] = before["ret"]
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_update_comment(
            resource_type="aws.guardduty.organization_configuration",
            name=resource_id,
        )
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={},
            desired_state={
                "name": name,
                "auto_enable": auto_enable,
                "resource_id": resource_id,
            },
        )
        return result

    ret = await hub.exec.aws.guardduty.organization_configuration.update(
        ctx,
        resource_id=resource_id,
        auto_enable=auto_enable,
        data_sources=data_sources,
        org_conf=before,
    )

    result["result"] = ret["result"]
    result["comment"] = ret["comment"]
    if not result["result"]:
        return result

    after = await hub.exec.aws.guardduty.organization_configuration.get(
        ctx, resource_id=resource_id
    )
    result["new_state"] = after["ret"]
    return result


async def absent(hub, ctx, name: str) -> Dict[str, Any]:
    """A No-Op function for aws.guardduty.organization_configuration.

    This is a configuration resource of the aws.guardduty.organization_configuration resource.
    It's not possible to delete aws.guardduty.organization_configuration.
    If you want to modify the aws.guardduty.organization_configuration resource,
    use the aws.guardduty.organization_configuration.present.

    Args:
        name:
            An Idem name of the resource.

    Request Syntax:
        .. code-block:: sls

            [guardduty.organization_configuration_id]:
              aws.guardduty.organization_configuration.absent:
                - name: "string"

    Returns:
        Dict[str, Any]
    """
    result = dict(
        comment=(
            "No-Op: The aws.guardduty.organization_configuration can not be deleted",
        ),
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem describe aws.guardduty.organization_configuration
    """
    result = {}

    ret = await hub.exec.aws.guardduty.detector.list(ctx, name="list detectors")

    if not ret["result"]:
        hub.log.debug(f"Could not list detector {ret['comment']}")
        return {}

    for detector in ret["ret"]:
        response = await hub.exec.aws.guardduty.organization_configuration.get(
            ctx, resource_id=detector.get("resource_id")
        )
        if not response["result"]:
            hub.log.warning(
                f"Could not list organization_configuration {ret['comment']}"
            )
        else:
            result[detector.get("resource_id")] = {
                "aws.guardduty.organization_configuration.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in response["ret"].items()
                ]
            }
    return result



"""State module to manage AWS SES event destination."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    configuration_set_name: str,
    event_destination: make_dataclass(
        "EventDestinationDefinition",
        [
            ("Enabled", bool, field(default=None)),
            ("MatchingEventTypes", List[str], field(default=None)),
            (
                "KinesisFirehoseDestination",
                make_dataclass(
                    "KinesisFirehoseDestination",
                    [("IamRoleArn", str), ("DeliveryStreamArn", str)],
                ),
                field(default=None),
            ),
            (
                "CloudWatchDestination",
                make_dataclass(
                    "CloudWatchDestination",
                    [
                        (
                            "DimensionConfigurations",
                            List[
                                make_dataclass(
                                    "CloudWatchDimensionConfiguration",
                                    [
                                        ("DimensionName", str),
                                        ("DimensionValueSource", str),
                                        ("DefaultDimensionValue", str),
                                    ],
                                )
                            ],
                        )
                    ],
                ),
                field(default=None),
            ),
            (
                "SnsDestination",
                make_dataclass("SnsDestination", [("TopicArn", str)]),
                field(default=None),
            ),
            (
                "PinpointDestination",
                make_dataclass(
                    "PinpointDestination",
                    [("ApplicationArn", str, field(default=None))],
                ),
                field(default=None),
            ),
        ],
    ),
    resource_id: str = None,
) -> Dict[str, Any]:
    """Create or update an event destination.

    Events include message sends, deliveries, opens, clicks, bounces, and complaints.
    Event destinations are places that you can send information about these events to. For example, you can send
    event data to Amazon SNS to receive notifications when you receive bounces or complaints, or you can use Amazon
    Kinesis Data Firehose to stream data to Amazon S3 for long-term storage. A single configuration set can include
    more than one event destination.

    Args:
        resource_id(str, Optional):
            An identifier that identifies the event destination within the configuration set. Defaults to None.
        name(str):
            A name that identifies the event destination within the configuration set.
        configuration_set_name(str):
            The name of the configuration set.
        event_destination(dict[str, Any]):
            An object that defines the event destination.

            * Enabled (bool, Optional):
                If true, the event destination is enabled. When the event destination is enabled, the specified
                event types are sent to the destinations in this EventDestinationDefinition. If false, the event
                destination is disabled. When the event destination is disabled, events aren't sent to the
                specified destinations.

            * MatchingEventTypes (list[str], Optional):
                An array that specifies which events the Amazon SES API v2 should send to the destinations in
                this EventDestinationDefinition.

            * KinesisFirehoseDestination (dict[str, Any], Optional):
                An object that defines an Amazon Kinesis Data Firehose destination for email events. You can use
                Amazon Kinesis Data Firehose to stream data to other services, such as Amazon S3 and Amazon
                Redshift.

                * IamRoleArn (str):
                    The Amazon Resource Name (ARN) of the IAM role that the Amazon SES API v2 uses to send email
                    events to the Amazon Kinesis Data Firehose stream.
                * DeliveryStreamArn (str):
                    The Amazon Resource Name (ARN) of the Amazon Kinesis Data Firehose stream that the Amazon SES
                    API v2 sends email events to.

            * CloudWatchDestination (dict[str, Any], Optional):
                An object that defines an Amazon CloudWatch destination for email events. You can use Amazon
                CloudWatch to monitor and gain insights on your email sending metrics.

                * DimensionConfigurations (list[dict[str, Any]]):
                    An array of objects that define the dimensions to use when you send email events to Amazon
                    CloudWatch.

                    * DimensionName (str):
                        The name of an Amazon CloudWatch dimension associated with an email sending metric. The name has
                        to meet the following criteria:   It can only contain ASCII letters (a–z, A–Z), numbers (0–9),
                        underscores (_), or dashes (-).   It can contain no more than 256 characters.
                    * DimensionValueSource (str):
                        The location where the Amazon SES API v2 finds the value of a dimension to publish to Amazon
                        CloudWatch. To use the message tags that you specify using an X-SES-MESSAGE-TAGS header or a
                        parameter to the SendEmail or SendRawEmail API, choose messageTag. To use your own email
                        headers, choose emailHeader. To use link tags, choose linkTags.
                    * DefaultDimensionValue (str):
                        The default value of the dimension that is published to Amazon CloudWatch if you don't provide
                        the value of the dimension when you send an email. This value has to meet the following
                        criteria:   It can only contain ASCII letters (a–z, A–Z), numbers (0–9), underscores (_), or
                        dashes (-).   It can contain no more than 256 characters.

            * SnsDestination (dict[str, Any], Optional):
                An object that defines an Amazon SNS destination for email events. You can use Amazon SNS to
                send notification when certain email events occur.

                * TopicArn (str):
                    The Amazon Resource Name (ARN) of the Amazon SNS topic to publish email events to. For more
                    information about Amazon SNS topics, see the Amazon SNS Developer Guide.

            * PinpointDestination (dict[str, Any], Optional):
                An object that defines an Amazon Pinpoint project destination for email events. You can send
                email event data to a Amazon Pinpoint project to view metrics using the Transactional Messaging
                dashboards that are built in to Amazon Pinpoint. For more information, see Transactional
                Messaging Charts in the Amazon Pinpoint User Guide.

                * ApplicationArn (str, Optional):
                    The Amazon Resource Name (ARN) of the Amazon Pinpoint project to send email events to.

    Returns:
        .. code-block:: bash

            {"result": True|False, "comment": ("A tuple",), "ret": None|Dict}


    Examples:
        .. code-block:: sls

            resource_is_present:
              aws.sesv2.event_destination.present:
                - name: value
                - configuration_set_name: value
                - event_destination: value
    """
    result, before, changed, exists = (
        dict(comment=(), old_state=None, new_state=None, name=name, result=True),
        None,
        False,
        False,
    )
    if "SnsDestination" not in event_destination:
        result["comment"] = ("Only SnsDestination is supported now.",)
        result["result"] = False
        return result
    before = await hub.exec.aws.sesv2.event_destination.get(
        ctx=ctx,
        resource_id=name,
        configuration_set_name=configuration_set_name,
    )
    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result
    if before.get("ret"):
        exists = True
        result["old_state"] = before["ret"].copy()
        if result["old_state"]["event_destination"] != event_destination:
            changed = True
    payload = {
        "ConfigurationSetName": configuration_set_name,
        "EventDestinationName": name,
        "EventDestination": event_destination,
    }
    action, func = (
        ("update", "update_configuration_set_event_destination")
        if exists
        else ("create", "create_configuration_set_event_destination")
    )
    if ctx.get("test") and (not exists or changed):
        result[
            "new_state"
        ] = hub.tool.aws.sesv2.conversion_utils.convert_raw_event_destination_to_present(
            configuration_set_name=configuration_set_name,
            raw_resource=event_destination,
            name=name,
        )
        result["comment"] = getattr(
            hub.tool.aws.comment_utils, f"would_{action}_comment"
        )(resource_type="aws.sesv2.event_destination", name=name)
    elif exists and not changed:
        result["new_state"] = result["old_state"]
        result["comment"] = getattr(
            hub.tool.aws.comment_utils, "already_exists_comment"
        )(resource_type="aws.sesv2.event_destination", name=name)
    else:
        try:
            ret = await getattr(hub.exec.boto3.client.sesv2, func)(ctx, **payload)
            if not ret["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            else:
                result[
                    "new_state"
                ] = hub.tool.aws.sesv2.conversion_utils.convert_raw_event_destination_to_present(
                    configuration_set_name=configuration_set_name,
                    name=name,
                    raw_resource=event_destination,
                )
                result["comment"] = getattr(
                    hub.tool.aws.comment_utils, f"{action}_comment"
                )(resource_type="aws.sesv2.event_destination", name=name)
        except Exception as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, configuration_set_name: str = None, resource_id: str = None
) -> Dict[str, Any]:
    """Delete an event destination.

    Events include message sends, deliveries, opens, clicks, bounces, and complaints.
    Event destinations are places that you can send information about these events to. For example, you can send
    event data to Amazon SNS to receive notifications when you receive bounces or complaints, or you can use Amazon
    Kinesis Data Firehose to stream data to Amazon S3 for long-term storage.

    Args:
        name(str):
            The name of the event destination to delete.
        configuration_set_name(str, Optional):
            The name of the configuration set that contains the event destination to delete.
        resource_id(str, Optional):
            An identifier of the resource in the provider.

    Returns:
        .. code-block:: bash

            {"result": True|False, "comment": ("A tuple",), "ret": None|Dict}


    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.sesv2.event_destination.absent:
                - name: value
                - resource_id: value
                - configuration_set_name: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.sesv2.event_destination", name=name
        )
        return result

    if not configuration_set_name:
        result["comment"] = hub.tool.aws.comment_utils.missing_args_for_absent_comment(
            resource_type="aws.sesv2.event_destination",
            name=name,
            args=["configuration_set_name"],
        )
        result["result"] = False
        return result

    before = await hub.exec.aws.sesv2.event_destination.get(
        ctx=ctx,
        resource_id=resource_id,
        configuration_set_name=configuration_set_name,
    )

    if not before["result"]:
        result["comment"] = before["comment"]
        result["result"] = False
        return result

    if not before.get("ret"):
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.sesv2.event_destination", name=name
        )
        return result

    result["old_state"] = before.get("ret")

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.sesv2.event_destination", name=name
        )
        return result
    else:
        try:
            ret = await hub.exec.boto3.client.sesv2.delete_configuration_set_event_destination(
                ctx,
                **{
                    "ConfigurationSetName": configuration_set_name,
                    "EventDestinationName": resource_id,
                },
            )
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.sesv2.event_destination", name=name
            )
        except Exception as e:
            result["result"] = False
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            idem describe aws.sesv2.event_destination
    """
    result = await hub.exec.aws.sesv2.event_destination.list(ctx)

    if not result["result"]:
        hub.log.debug(f"Could not describe Event destination {result['comment']}")
        return {}

    return result["ret"]



"""Exec module for managing Guardduty detectors."""
__func_alias__ = {"list_": "list"}

from typing import Dict


async def get(hub, ctx, resource_id: str, name: str = None) -> Dict:
    """Get info about AWS Guardduty detector based on the detector_id passed.

    Args:
        resource_id(str):
            AWS Guardduty Detector id
        name(str, Optional):
            Name of the Idem state
    """
    result = dict(comment=[], ret=None, result=True)
    before = await hub.exec.boto3.client.guardduty.get_detector(
        ctx, DetectorId=resource_id
    )
    if not before["result"]:
        if "BadRequestException" in str(before["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.guardduty.detector", name=name
                )
            )
            return result
        result["comment"] += list(before["comment"])
        result["result"] = False
        return result

    result[
        "ret"
    ] = await hub.tool.aws.guardduty.conversion_utils.convert_raw_detector_to_present_async(
        ctx, raw_resource=before, idem_resource_name=resource_id
    )
    return result


async def list_(hub, ctx, name: str = None) -> Dict:
    """List AWS guard duty detectors.

    Args:
        name(str, Optional):
            Name of the Idem state for logging purposes.

    Returns:
        Dict[str, Any]:
            Return detectors.
    """
    result = dict(comment=[], ret=[], result=True)
    before = await hub.exec.boto3.client.guardduty.list_detectors(ctx)
    if not before["result"]:
        result["comment"] += list(before["comment"])
        result["result"] = False
        return result
    if not before["ret"]["DetectorIds"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.guardduty.detector", name=name
            )
        )
        return result
    for detector_id in before["ret"]["DetectorIds"]:
        detector_ret = await get(hub, ctx, resource_id=detector_id)
        if not detector_ret["result"]:
            hub.log.warning(
                f"Could not get detector info for id {detector_id}, hence skipping it in list"
            )
            result["comment"].append(detector_ret["comment"])
            continue
        result["ret"].append(detector_ret["ret"])

    return result



from typing import Dict

import boto3


async def gather(hub, profiles):
    """
    Load profiles from AWS credential files

    If no aws profiles were found in the acct_file, profiles will be extrapolated from awscli credentials.

    Example:
    .. code-block:: yaml

        aws:
          profile_name:
            region_name: us-west-1
            endpoint_url: localhost:992
            aws_access_key_id: my_key_id
            aws_secret_access_key: my_key
            aws_session_token: my_token
    """
    sub_profiles = {}

    # Use the profiles from an acct_file
    if profiles:
        for profile, ctx in profiles.get("aws", {}).items():
            sub_profiles[profile] = ctx
    # If there were no aws profiles in the acct_file, fallback to credentials from awscli
    else:
        default_idem_profile = hub.OPT.idem.get("acct_profile", hub.acct.DEFAULT)
        # Create a new session with the credentials boto3 uses by default (awscli)
        session = boto3.Session()

        # Add profiles based on the profile names available in awscli/creds
        for profile in session.available_profiles:
            new_profile = _profile_from_session(boto3.Session(profile_name=profile))
            if new_profile:
                sub_profiles[profile] = new_profile

        # Create a default profile
        if default_idem_profile not in sub_profiles:
            # Get the credentials for the default profile
            new_profile = _profile_from_session(session)
            if new_profile:
                sub_profiles[default_idem_profile] = new_profile

    return sub_profiles


def _profile_from_session(session: boto3.Session) -> Dict[str, str]:
    """
    Get credentials from a boto3 session and return an acct profile
    """
    result = {}
    credentials = session.get_credentials()
    if credentials:
        # Put the credentials into a plain dictionary
        if credentials.access_key:
            result["aws_access_key_id"] = credentials.access_key
        if credentials.secret_key:
            result["aws_secret_access_key"] = credentials.secret_key
        if credentials.token:
            result["aws_session_token"] = credentials.token
    return result



from typing import Dict
from typing import List


async def search_raw(
    hub,
    ctx,
    resource_id: str = None,
    filters: List = None,
    tags: List = None,
    executable_users: List = None,
    owners: List = None,
    include_deprecated: bool = False,
) -> Dict:
    """Search EC2 ami from AWS account.

    Fetch one or more ami's from AWS. The return will be in the same format as what the boto3 api returns.

    Args:
        resource_id(str, Optional):
            Resource ID of the ami.

        executable_users (list[str], Optional):
            Scopes the images by users with explicit launch permissions.
            Specify an Amazon Web Services account ID, self (the sender of the request), or all (public AMIs).

        owners (list[str], Optional):
            Scopes the results to images with the specified owners.
            You can specify a combination of Amazon Web Services account IDs, self , amazon , and aws-marketplace.
            If you omit this parameter, the results include all images for which you have launch permissions, regardless of ownership.

        include_deprecated (bool, Optional):
            If true , all deprecated AMIs are included in the response.
            If false , no deprecated AMIs are included in the response. If no value is specified, the default value is false .

        filters(List, Optional):
            One or more filters: for example, tag :<key>, tag-key. A complete list of filters can be found at
            https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.describe_images

        tags(List, Optional):
            The list of tags to filter by. For example, to find all resources that have a tag with the key
            "Owner" and the value "TeamA" , specify "tag:Owner" for the Dict key and "TeamA" for the Dict value.

    Returns:
        {"result": True|False, "comment": A message List, "ret": Dict}

    """
    result = dict(comment=[], ret=None, result=True)
    syntax_validation = hub.tool.aws.search_utils.search_filter_syntax_validation(
        filters=filters
    )
    if not syntax_validation["result"]:
        result["comment"] = list(syntax_validation["comment"])
        return result
    boto3_filter = hub.tool.aws.search_utils.convert_search_filter_to_boto3(
        filters=filters
    )
    if tags is not None:
        for tag in tags:
            boto3_filter.append({"Name": tag["Key"], "Values": [tag["Value"]]})

    ami_args = {}
    if filters:
        ami_args["Filters"] = boto3_filter
    if resource_id:
        ami_args["ImageIds"] = [resource_id]
    if executable_users:
        ami_args["ExecutableUsers"] = executable_users
    if owners:
        ami_args["Owners"] = owners
    if include_deprecated:
        ami_args["IncludeDeprecated"] = include_deprecated

    ret = await hub.exec.boto3.client.ec2.describe_images(ctx, **ami_args)

    result["result"] = ret["result"]
    result["comment"] = list(ret["comment"])
    result["ret"] = ret["ret"]
    return result



"""Utility functions for EC2 VPC Endpoint Service Configurations."""
from typing import Any
from typing import Dict


async def convert_raw_resource_to_present_async(
    hub, ctx, idem_resource_name: str, resource_id: str, raw_resource: dict
) -> Dict[str, Any]:
    r"""
    Convert raw resource of vpc_endpoint_service_configuration type into present format.
    """
    resource_translated = {"resource_id": resource_id}

    resource_parameters = {
        "ServiceType": "service_type",
        "ServiceId": "service_id",
        "ServiceName": "service_name",
        "ServiceState": "service_state",
        "AvailabilityZones": "availability_zones",
        "AcceptanceRequired": "acceptance_required",
        "ManagesVpcEndpoints": "manages_vpc_endpoints",
        "NetworkLoadBalancerArns": "network_load_balancer_arns",
        "GatewayLoadBalancerArns": "gateway_load_balancer_arns",
        "SupportedIpAddressTypes": "supported_ip_address_types",
        "BaseEndpointDnsNames": "base_endpoint_dns_names",
        "PrivateDnsName": "private_dns_name",
        "PrivateDnsNameConfiguration": "private_dns_name_configuration",
        "PayerResponsibility": "payer_responsibility",
    }

    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource and raw_resource.get(parameter_raw):
            resource_translated[parameter_present] = raw_resource.get(parameter_raw)

    # The resource name could be given by the input or auto generated
    # Default idem creation adds name to tags with Name key (as specified in AWS console)
    resource_name = (
        raw_resource.get("tags", {}).get("Name")
        or idem_resource_name
        or raw_resource.get("ServiceName")
    )
    resource_translated["name"] = resource_name

    resource_translated["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
        raw_resource.get("Tags")
    )

    return resource_translated


def evaluate_update_desired_state(
    hub, ctx, current_state: dict, desired_state: dict
) -> Dict[str, Any]:
    r"""
    Evaluates desired state for updating vpc_endpoint_service_configuration
    """
    (
        is_config_equal,
        network_lb_added,
        network_lb_removed,
    ) = _evaluate_added_removed_config(
        "network_load_balancer_arns", current_state, desired_state
    )
    if not is_config_equal:
        desired_state["add_network_load_balancer_arns"] = network_lb_added
        desired_state["remove_network_load_balancer_arns"] = network_lb_removed
    (
        is_config_equal,
        gateway_lb_added,
        gateway_lb_removed,
    ) = _evaluate_added_removed_config(
        "gateway_load_balancer_arns", current_state, desired_state
    )
    if not is_config_equal:
        desired_state["add_gateway_load_balancer_arns"] = gateway_lb_added
        desired_state["remove_gateway_load_balancer_arns"] = gateway_lb_removed
    (
        is_config_equal,
        ip_address_types_added,
        ip_address_types_removed,
    ) = _evaluate_added_removed_config(
        "supported_ip_address_types", current_state, desired_state
    )
    if not is_config_equal:
        desired_state["add_supported_ip_address_types"] = ip_address_types_added
        desired_state["remove_supported_ip_address_types"] = ip_address_types_removed

    # Default to False. If desired state no longer wants to have private dns name, we can switch it to True.
    desired_state["remove_private_dns_name"] = False
    if (
        current_state.get("private_dns_name", None) is not None
        and desired_state.get("private_dns_name") is None
    ):
        desired_state["remove_private_dns_name"] = True

    return desired_state


def _evaluate_added_removed_config(
    config_name: str, current_state: dict, desired_state: dict
):
    current = current_state.get(config_name, None)
    desired = desired_state.get(config_name, None)

    is_config_equal = current == desired
    added = list(set(desired) - set(current)) if desired else None
    removed = list(set(current) - set(desired)) if current else None

    if config_name in desired_state:
        del desired_state[config_name]

    return is_config_equal, added, removed



"""Exec module for managing organization accounts."""
__func_alias__ = {"list_": "list"}

from typing import Dict, Any


async def get(hub, ctx, resource_id: str, name: str = None) -> Dict[str, Any]:
    """Get info about AWS Organizations account based on resource_id.

    Args:
        resource_id(str):
            AWS Organizations account id

        name(str, Optional):
            Name of the Idem state

    Returns:
        Dict[str, Any]:
            Returns account in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.organizations.account.get name="get account" resource_id="account_id"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resources:
              exec.run:
                - path: aws.organizations.account.get
                - kwargs:
                    name: "get account info"
                    resource_id: "257230699585"
    """
    result = dict(result=True, ret=None, comment=[])
    ret = await hub.exec.boto3.client.organizations.describe_account(
        ctx, AccountId=resource_id
    )

    if not ret["result"]:
        if "AccountNotFoundException" in str(ret["comment"]):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.organizations.account", name=name
                )
            )
            return result
        result["result"] = ret["result"]
        result["comment"] += list(ret["comment"])
        return result

    account = ret["ret"]["Account"]
    if not account:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.organizations.account", name=name
            )
        )
        return result
    # get tags
    tags_ret = await hub.exec.boto3.client.organizations.list_tags_for_resource(
        ctx, ResourceId=resource_id
    )
    if not tags_ret["result"]:
        result["result"] = False
        result["comment"] += list(tags_ret["comment"])
        return result

    tags = None
    if tags_ret["ret"]["Tags"]:
        tags = tags_ret["ret"]["Tags"]

    # get parent_id
    parents_ret = await hub.exec.boto3.client.organizations.list_parents(
        ctx, ChildId=resource_id
    )
    if not parents_ret["result"]:
        result["result"] = False
        result["comment"] += list(parents_ret["comment"])
        return result

    current_parent_id = None
    if parents_ret["ret"]["Parents"]:
        current_parent_id = parents_ret["ret"]["Parents"][0]["Id"]

    result[
        "ret"
    ] = hub.tool.aws.organizations.conversion_utils.convert_raw_account_to_present(
        parent_id=current_parent_id, tags=tags, account=account
    )

    return result


async def list_(hub, ctx, name: str = None) -> Dict[str, Any]:
    """List AWS accounts in an organization.

    Args:
        name(str, Optional):
            Name of the Idem state.

    Returns:
        Dict[str, Any]:
            Returns organization accounts in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec aws.organizations.account.list name="list accounts"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resources:
              exec.run:
                - path: aws.organizations.account.list
                - kwargs:
                    name: my_accounts
    """

    result = dict(result=True, comment=[], ret=[])
    ret = await hub.exec.boto3.client.organizations.list_accounts(ctx)
    if not ret["result"]:
        result["result"] = ret["result"]
        result["comment"] += list(ret["comment"])
        return result

    if not ret["ret"]["Accounts"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.list_empty_comment(
                resource_type="aws.organizations.account", name=name
            )
        )
        return result
    for account in ret["ret"]["Accounts"]:
        account_id = account.get("Id")
        account_ret = await get(hub, ctx, resource_id=account_id)
        if not account_ret["result"]:
            hub.log.debug(
                f"Unable to get info for account: {account_id} Error:  {account_ret['comment']}. Hence skipping it in list"
            )
            continue
        result["ret"].append(account_ret["ret"])
    return result



"""Exec module for Amazon Configuration Aggregator."""
from typing import Dict
from typing import List

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str) -> Dict:
    """Get a Configuration Aggregator resource from AWS for a given name or resource_id.

    Args:
        name(str):
            The name of the Configuration Aggregator.

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.config.configuration_aggregator.get name="name"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.config.configuration_aggregator.get
                - kwargs:
                    name: "unmanaged_configuration_aggregator"
    """
    result = dict(comment=[], ret=None, result=True)
    configuration_aggregator = (
        await hub.exec.boto3.client.config.describe_configuration_aggregators(
            ctx, ConfigurationAggregatorNames=[name]
        )
    )
    if not configuration_aggregator["result"]:
        if "NoSuchConfigurationAggregatorException" in str(
            configuration_aggregator["comment"]
        ):
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.config.configuration_aggregator", name=name
                )
            )
            return result
        result["comment"] += list(configuration_aggregator["comment"])
        result["result"] = False
        return result
    if not configuration_aggregator["ret"]["ConfigurationAggregators"]:
        result["comment"].append(
            hub.tool.aws.comment_utils.get_empty_comment(
                resource_type="aws.config.configuration_aggregator", name=name
            )
        )
        return result

    ret = hub.tool.aws.config.conversion_utils.convert_raw_config_aggregator_to_present(
        ctx,
        raw_resource=((configuration_aggregator["ret"])["ConfigurationAggregators"])[0],
    )
    result["ret"] = ret
    return result


async def list_(hub, ctx, names: List[str] = None) -> Dict:
    """Get a Configuration Aggregator resources from AWS.

    Args:
        names(List[str] = Optional):
            The name of the Configuration Aggregators.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/config.html#ConfigService.Client.describe_configuration_aggregators

    Examples:
        Calling from the CLI:

        .. code-block:: bash

            $ idem exec aws.config.configuration_aggregator.list configuration_aggregator_names="names"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: aws.config.configuration_aggregator.list
                - kwargs:
                    names: ["unmanaged_configuration_aggregator"]
    """
    result = dict(comment=[], ret=[], result=True)
    if names:
        for name in names:
            configuration_aggregator_ret = await get(hub, ctx, name)
            if not configuration_aggregator_ret["result"]:
                hub.log.warning(
                    f"Could not get aggregator info for name {name}, hence skipping it in list"
                )
                result["comment"].append(configuration_aggregator_ret["comment"])
                continue
            result["ret"].append(configuration_aggregator_ret["ret"])
    else:
        configuration_aggregators = (
            await hub.exec.boto3.client.config.describe_configuration_aggregators(ctx)
        )
        if not configuration_aggregators["result"]:
            result["comment"] = configuration_aggregators["comment"]
            result["result"] = False
            return result

        if not configuration_aggregators["ret"]["ConfigurationAggregators"]:
            result["comment"].append(
                hub.tool.aws.comment_utils.get_empty_comment(
                    resource_type="aws.config.configuration_aggregator",
                    name="configuration aggregators",
                )
            )
            return result
        for configuration_aggregator in configuration_aggregators["ret"][
            "ConfigurationAggregators"
        ]:
            resource_translated = hub.tool.aws.config.conversion_utils.convert_raw_config_aggregator_to_present(
                ctx, raw_resource=configuration_aggregator
            )
            result["ret"].append(resource_translated)
    return result



"""State module for managing AWS S3 bucket encryption."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

STATE_NAME = "aws.s3.bucket_encryption"


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    bucket: str,
    server_side_encryption_configuration: make_dataclass(
        "ServerSideEncryptionConfiguration",
        [
            (
                "Rules",
                List[
                    make_dataclass(
                        "ServerSideEncryptionRule",
                        [
                            (
                                "ApplyServerSideEncryptionByDefault",
                                make_dataclass(
                                    "ServerSideEncryptionByDefault",
                                    [
                                        ("SSEAlgorithm", str),
                                        ("KMSMasterKeyID", str, field(default=None)),
                                    ],
                                ),
                                field(default=None),
                            ),
                            ("BucketKeyEnabled", bool, field(default=False)),
                        ],
                    )
                ],
            )
        ],
    ),
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=4)),
                        ("max_attempts", int, field(default=30)),
                    ],
                ),
                field(default=None),
            )
        ],
    ) = None,
) -> Dict[str, Any]:
    """Creates an encryption configuration for an S3 bucket resource.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            The name of the S3 bucket in Amazon Web Services. It must be equal to the bucket parameter.

        bucket(str):
            The name of the S3 bucket in Amazon Web Services.

        server_side_encryption_configuration(Dict[str, Any]):
            Specifies the default server-side-encryption configuration.

            * Rules (list[Dict[str, Any]]):
                Container for information about a particular server-side encryption configuration rule.

                * ApplyServerSideEncryptionByDefault (Dict[str, Any], Optional):
                    Specifies the default server-side encryption to apply to new objects in the bucket. If a PUT
                    Object request doesn't specify any server-side encryption, this default encryption will be
                    applied.

                    * SSEAlgorithm (str):
                        Server-side encryption algorithm to use for the default encryption.

                    * KMSMasterKeyID (str, Optional):
                        Amazon Web Services Key Management Service (KMS) customer Amazon Web Services KMS key ID to use
                        for the default encryption. This parameter is allowed if and only if SSEAlgorithm is set to
                        aws:kms. You can specify the key ID or the Amazon Resource Name (ARN) of the KMS key. If you use
                        a key ID, you can run into a LogDestination undeliverable error when creating a VPC flow log.
                        If you are using encryption with cross-account or Amazon Web Services service operations you
                        must use a fully qualified KMS key ARN. For more information, see Using encryption for cross-
                        account operations.    Key ID: 1234abcd-12ab-34cd-56ef-1234567890ab    Key ARN: arn:aws:kms:us-
                        east-2:111122223333:key/1234abcd-12ab-34cd-56ef-1234567890ab     Amazon S3 only supports
                        symmetric encryption KMS keys. For more information, see Asymmetric keys in Amazon Web Services
                        KMS in the Amazon Web Services Key Management Service Developer Guide.

                * BucketKeyEnabled (bool, Optional):
                    Specifies whether Amazon S3 should use an S3 Bucket Key with server-side encryption using KMS
                    (SSE-KMS) for new objects in the bucket. Existing objects are not affected. Setting the
                    BucketKeyEnabled element to true causes Amazon S3 to use an S3 Bucket Key. By default, S3 Bucket
                    Key is not enabled. For more information, see Amazon S3 Bucket Keys in the Amazon S3 User Guide.

                    Defaults to False.

        timeout(dict, Optional):
            Timeout configuration for S3 bucket encryption configuration.

            * update (str):
                Timeout configuration for updating the S3 bucket encryption configuration.

                * delay (int, Optional):
                    The amount of time in seconds to wait between attempts. Defaults to 4 seconds.

                * max_attempts (int, Optional):
                    Maximum attempts of waiting for the update. Defaults to 30 attempts.

    Request Syntax:
        .. code-block:: yaml

            [idem_test_aws_s3_bucket_encryption]:
              aws.s3.bucket_encryption.present:
                - name: 'string'
                - bucket: 'string'
                - server_side_encryption_configuration:
                    Rules:
                    - ApplyServerSideEncryptionByDefault:
                        SSEAlgorithm: 'string'
                        KMSMasterKeyID: 'string'
                      BucketKeyEnabled: True|False

    Returns:
        dict[str, Any]

    Examples:
        .. code-block:: yaml

            idem_test_aws_s3_bucket_encryption:
              aws.s3.bucket_encryption.present:
                - name: value
                - bucket: value
                - server_side_encryption_configuration:
                    Rules:
                    - ApplyServerSideEncryptionByDefault:
                        SSEAlgorithm: 'AES256'
                        KMSMasterKeyID: 'string'
                      BucketKeyEnabled: True
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    before_ret = None

    if resource_id:
        if resource_id != bucket:
            result["result"] = False
            result["comment"] = (
                f"Bucket '{bucket}' and resource_id '{resource_id}' parameters must be the same",
            )
            return result

        before_ret = await hub.exec.boto3.client.s3.get_bucket_encryption(
            ctx, Bucket=resource_id
        )
        if not before_ret["result"]:
            result["result"] = False
            result["comment"] = before_ret["comment"]
            return result

    if before_ret:
        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type=STATE_NAME, name=name
        )
        result[
            "old_state"
        ] = hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
            bucket=resource_id,
            raw_resource=before_ret.get("ret"),
            idem_resource_name=name,
        )
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "server_side_encryption_configuration": server_side_encryption_configuration,
        }

        if hub.tool.aws.s3.bucket_encryption.is_bucket_encryption_configuration_identical(
            resource_parameters, result["old_state"]
        ):
            # no updates to apply
            return result

        if ctx.get("test", False):
            result["new_state"].update(resource_parameters)
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.would_update_comment(
                resource_type=STATE_NAME, name=name
            )
            return result
        else:
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.update_comment(
                resource_type=STATE_NAME, name=name
            )
    else:
        if ctx.get("test", False):
            result[
                "new_state"
            ] = raw_resource = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "bucket": bucket,
                    "server_side_encryption_configuration": server_side_encryption_configuration,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type=STATE_NAME, name=name
            )
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type=STATE_NAME, name=name
        )

    put_ret = await hub.exec.boto3.client.s3.put_bucket_encryption(
        ctx,
        Bucket=bucket,
        ServerSideEncryptionConfiguration=server_side_encryption_configuration,
    )
    if not put_ret["result"]:
        result["result"] = False
        result["comment"] = result["comment"] + put_ret["comment"]
        return result

    wait_for_updates_ret = await hub.tool.aws.s3.bucket_encryption.wait_for_updates(
        ctx,
        bucket=bucket,
        server_side_encryption_configuration=server_side_encryption_configuration,
        timeout=timeout.get("update") if timeout else None,
    )
    if not wait_for_updates_ret["result"]:
        result["result"] = False
        result["comment"] += wait_for_updates_ret["comment"]
        return result

    result[
        "new_state"
    ] = hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
        bucket=bucket,
        raw_resource=wait_for_updates_ret.get("ret"),
        idem_resource_name=name,
    )

    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, timeout: Dict = None
) -> Dict[str, Any]:
    """Deletes an encryption configuration from an S3 bucket resource.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            The name of the S3 bucket in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

        timeout(dict, Optional):
            Timeout configuration for S3 bucket encryption configuration.

            * delete (str):
                Timeout configuration for deleting the S3 bucket encryption configuration.

                * delay (int, Optional):
                    The amount of time in seconds to wait between attempts. Defaults to 4 seconds.

                * max_attempts (int, Optional):
                    Maximum attempts of waiting for the deletion. Defaults to 30 attempts.

    Returns:
        dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_s3_bucket_encryption:
              aws.s3.bucket_encryption.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=STATE_NAME, name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.s3.get_bucket_encryption(
        ctx, Bucket=resource_id
    )
    if not before_ret["result"]:
        if "ServerSideEncryptionConfigurationNotFoundError" in str(
            before_ret["comment"][0]
        ) or "NoSuchBucket" in str(before_ret["comment"][0]):
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type=STATE_NAME, name=name
            )
        else:
            result["result"] = False
            result["comment"] = before_ret["comment"]
        return result

    result[
        "old_state"
    ] = hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
        bucket=resource_id, raw_resource=before_ret["ret"], idem_resource_name=name
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=STATE_NAME, name=name
        )
        return result

    # The delete_bucket_encryption would put back default encryption configuration
    # The default configuration: (`Server-side encryption with Amazon S3 managed keys (SSE-S3)`, `Bucket Key disabled`)
    delete_ret = await hub.exec.boto3.client.s3.delete_bucket_encryption(
        ctx, Bucket=resource_id
    )
    if not delete_ret["result"]:
        result["result"] = False
        result["comment"] = delete_ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type=STATE_NAME, name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Gets the encryption configuration for each S3 bucket under the given AWS account.

    Returns:
        dict[str, dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.s3.bucket_encryption
    """
    result = {}

    list_buckets_ret = await hub.exec.boto3.client.s3.list_buckets(ctx)
    if not list_buckets_ret["result"]:
        hub.log.debug(f"Could not list S3 buckets: {list_buckets_ret['comment']}")
        return result

    for bucket in list_buckets_ret["ret"]["Buckets"]:
        bucket_name = bucket.get("Name")

        get_bucket_encryption_ret = (
            await hub.exec.boto3.client.s3.get_bucket_encryption(
                ctx, Bucket=bucket_name
            )
        )
        if not get_bucket_encryption_ret["result"]:
            if "ServerSideEncryptionConfigurationNotFoundError" not in str(
                get_bucket_encryption_ret["comment"][0]
            ):
                hub.log.debug(
                    f"Could not get encryption configuration for S3 bucket '{bucket_name}': "
                    f"{get_bucket_encryption_ret['comment']}. Describe will skip this S3 bucket and continue."
                )
            continue

        resource_translated = (
            hub.tool.aws.s3.bucket_encryption.convert_raw_bucket_encryption_to_present(
                bucket=bucket_name,
                raw_resource=get_bucket_encryption_ret["ret"],
            )
        )

        result[resource_translated["name"]] = {
            f"{STATE_NAME}.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result
